"""
Classes for checking if a release violates the tracker's rules
"""

import abc

from ... import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerRuleBase(abc.ABC):
    """
    Abstract base class that checks a release against one rule

    :param tracker_jobs: Instance of a :class:`~.TrackerJobsBase` subclass
    """

    required_jobs = ()
    """
    Sequence of :class:`~.TrackerJobsBase` attribute names that resolve to :class:`~.JobBase`
    instances

    All required jobs must finish before :meth:`check` is called.
    """

    def __init__(self, tracker_jobs):
        self._tracker_jobs = tracker_jobs

    @property
    def tracker_jobs(self):
        """:class:`~.TrackerJobsBase` instance"""
        return self._tracker_jobs

    @property
    def release_name(self):
        """:class:`~.ReleaseName` instance"""
        return self._tracker_jobs.release_name

    @property
    def tracker(self):
        """:class:`~.TrackerBase` instance"""
        return self._tracker_jobs.tracker

    async def _wait_for_required_jobs(self):
        """Block until all :attr:`~.required_jobs` are finished"""
        jobs = tuple(
            getattr(self.tracker_jobs, job_name)
            for job_name in self.required_jobs
        )
        _log.debug('Rule %s: Waiting for required jobs: %s', type(self).__name__, tuple(j.name for j in jobs))
        for job in jobs:
            await job.wait_finished()
        _log.debug('Rule %s: Done waiting for required jobs: %s', type(self).__name__, tuple(j.name for j in jobs))

    async def check(self):
        """
        Wait for :attr:`required_jobs` and check if rule is broken

        The actual checking is done in :meth:`_check`, which must be implemented by the subclass and
        raise :class:`~.RuleBroken` if the rule is broken.
        """
        await self._wait_for_required_jobs()
        await self._check()

    @abc.abstractmethod
    async def _check(self):
        """Check if rule is broken"""


class BannedGroup(TrackerRuleBase):
    """Check if release group is not allowed"""

    banned_groups = set()
    """
    :class:`set` of banned group names

    Groups specified here are always banned. To ban groups conditionally (e.g. ban only encodes from
    a certain group), override :meth:`_check_custom`.
    """

    def is_group(self, group_name):
        """
        Return whether `group_name` is equal to the :attr:`~.ReleaseName.group` of
        :attr:`~.TrackerRuleBase.release_name`
        """
        return self.release_name.group.lower() == group_name.lower()

    async def _check(self):
        await self._check_custom()

        # Case-insensitively match group name against `banned_groups`.
        for banned_group in self.banned_groups:
            if self.is_group(banned_group):
                raise errors.BannedGroup(banned_group)

    async def _check_custom(self):
        """
        Called by :meth:`check` before simple group name matching is done

        This method should be implemented by subclasses to ban certain groups only in some cases.

        :raise RuleBroken: if the group is banned
        """


class HdOnly(TrackerRuleBase):
    """Check if release is HD"""

    message = 'Not a HD release'
    """Error message if release is not HD"""

    async def _check(self):
        if utils.mediainfo.video.get_resolution_int(self.release_name.path) < 720:
            raise errors.RuleBroken(self.message)



from ..base import rules


class AntBannedGroup(rules.BannedGroup):
    banned_groups = {
        '3LTON',
        '4yEo',
        'ADE',
        'AFG',
        'AniHLS',
        'AnimeRG',
        'AniURL',
        'AROMA',
        'aXXo',
        'Brrip',
        'CHD',
        'CM8',
        'CrEwSaDe',
        'd3g',
        'DDR',
        'DeadFish',
        'DNL',
        'ELiTE',
        'eSc',
        'FaNGDiNG0',
        'FGT',
        'Flights',
        'FRDS',
        'FUM',
        'HAiKU',
        'HD2DVD',
        'HDS',
        'HDTime',
        'Hi10',
        'ION10',
        'iPlanet',
        'JIVE',
        'KiNGDOM',
        'Leffe',
        'LiGaS',
        'LOAD',
        'MeGusta',
        'mHD',
        'MkvCage',
        'mSD',
        'NhaNc3',
        'nHD',
        'NOIVTC',
        'nSD',
        'Oj',
        'Ozlem',
        'PiRaTeS',
        'PRoDJi',
        'RAPiDCOWS',
        'RARBG',
        'RDN',
        'REsuRRecTioN',
        'RetroPeeps',
        'RMTeam',
        'SANTi',
        'SicFoI',
        'SPASM',
        'SPDVD',
        'STUTTERSHIT',
        'TBS',
        'Telly',
        'TM',
        'UPiNSMOKE',
        'URANiME',
        'WAF',
        'xRed',
        'XS',
        'YIFY',
        'YTS',
        'Zeus',
        'ZKBL',
        'ZmN',
        'ZMNT',
    }



"""
Abstract base class for all tracker-specific stuff
"""

import abc
import asyncio
import collections
import functools
import inspect
import os
import types

from ... import __project_name__, errors, utils
from ._howto import Howto
from .rules import TrackerRuleBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerBase(abc.ABC):
    """
    Base class for tracker-specific operations, e.g. uploading

    :param options: User configuration options for this tracker,
        e.g. authentication details, announce URL, etc
    :type options: :class:`dict`-like
    """

    @property
    @abc.abstractmethod
    def TrackerJobs(self):
        """Subclass of :class:`~.TrackerJobsBase`"""

    @property
    @abc.abstractmethod
    def TrackerConfig(self):
        """Subclass of :class:`~.TrackerConfigBase`"""

    rules = None
    """
    Sequence of :class:`~.TrackerRuleBase` subclasses or `None`

    For convenience, this may also be a :class:`~.ModuleType` that provides
    :class:`~.TrackerRuleBase` subclasses.
    """

    def __init_subclass__(cls):
        super().__init_subclass__()

        # Turn cls.rules into sequence of TrackerRileBase subclasses.
        if isinstance(cls.rules, types.ModuleType):
            cls.rules = utils.subclasses(TrackerRuleBase, modules=(cls.rules,))
        elif isinstance(cls.rules, collections.abc.Iterable):
            cls.rules = tuple(cls.rules)
        elif cls.rules is None:
            cls.rules = ()
        else:
            raise TypeError(f'Invalid {cls.__name__}.rules: {cls.rules!r}')

    def __init__(self, options=None):
        self._options = options or {}
        self._signal = utils.signal.Signal(
            id=f'{self.name}-tracker',
            signals=(
                'warning',
                'error',
                'exception',
                'logging_in', 'login_failed', 'logged_in',
                'logging_out', 'logout_failed', 'logged_out',
            ),
        )

    @property
    @abc.abstractmethod
    def name(self):
        """Lower-case tracker name abbreviation for internal use"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing tracker name abbreviation"""

    @property
    @abc.abstractmethod
    def torrent_source_field(self):
        """
        Torrents for this tracker get a ``source`` field with this value

        This is usually the same as :attr:`label`.
        """

    setup_howto_template = 'Nobody has written a setup howto yet.'
    """
    Step-by-step guide that explains how to make your first upload

    .. note:: This MUST be a class attribute and not a property.

    The following placeholders can be used in f-string format:

        - ``howto`` - :class:`~._howto.Howto` instance
        - ``tracker`` - :class:`~.TrackerBase` subclass
        - ``executable`` - Name of the executable that runs the application
    """

    @classmethod
    def generate_setup_howto(cls):
        """Fill in any placeholders in :attr:`setup_howto_template`"""
        return utils.string.evaluate_fstring(
            cls.setup_howto_template,
            howto=Howto(tracker_cls=cls),
            tracker=cls,
            executable=__project_name__,
        )

    @property
    def options(self):
        """
        Configuration options provided by the user

        This is the :class:`dict`-like object from the initialization argument
        of the same name.
        """
        return self._options

    @functools.cached_property
    def _tasks(self):
        return []

    def attach_task(self, coro, callback=None):
        """
        Run awaitable `coro` in background task

        :param coro: Any awaitable
        :param callback: Callable that is called with the task after `coro`
            returned
        """
        def callback_(task):
            self._tasks.remove(task)
            if callback:
                callback(task)

        task = utils.run_task(coro, callback=callback_)
        self._tasks.append(task)
        _log.debug('%s: Attached task: %r', self.name, task)
        return task

    async def await_tasks(self):
        """Wait for all awaitables passed to :meth:`attach_task`"""
        for task in self._tasks:
            await task

    async def login(self, **credentials):
        """
        End user session by calling :attr:`_login`, which must be implemented by the subclass

        Set :attr:`is_logged_in` to `True` or `False`. Emit `logged_in` :attr:`signal` on success.

        Any keyword arguments are passed on to :meth:`_login`, e.g. to provide a 2FA one-time
        password.

        :meth:`_login` should raise :class:`~.TfaRequired` to get called again with a `tfa_otp`
        keyword argument. The `tfa_otp` value will be from a user prompt that asks for a 2FA
        one-time password. (See :meth:`~.TrackerJobsBase.login_job`.)

        :raise errors.RequestError: on failure
        :raise errors.TfaRequired: if login failed due to missing second factor
        """
        async with self._session_lock:
            # Check internal flag.
            if not self.is_logged_in:
                self.signal.emit('logging_in')
                try:
                    await self._login(**credentials)
                    await self.confirm_logged_in()
                except errors.RequestError as e:
                    _log.debug('Login failed: %r', e)
                    self._is_logged_in = False
                    self.signal.emit('login_failed', e)
                    raise
                else:
                    _log.debug('Logged in successfully')
                    self._is_logged_in = True
                    self.signal.emit('logged_in')

    async def still_logged_in(self):
        """
        Return whether we have a valid user session stored in :attr:`cookies_filepath`

        If no :attr:`cookies_filepath` is specified, return `None`.

        Otherwise, catch :class:`~.RequestError` from :meth:`confirm_logged_in` to check if we are
        logged in.

        If we are not logged in, call :meth:`delete_cookies_filepath`, emit the ``login_failed``
        :attr:`~.signal` and return `False`.

        If we are logged in, emit the ``logged_in`` :attr:`~.signal` and return `True`.
        """
        if self.cookies_filepath:
            try:
                await self.confirm_logged_in()
            except errors.RequestError as e:
                _log.debug('Failed to validate stored user session: %r: %r', self.cookies_filepath, e)
                self.delete_cookies_filepath()
                self._is_logged_in = False
                self.signal.emit('login_failed', f'Session is no longer valid: {e}')
                return False
            else:
                _log.debug('User session is still valid: %r', self.cookies_filepath)
                self._is_logged_in = True
                self.signal.emit('logged_in')
                return True
        else:
            _log.debug('No user session was stored: %r', self.cookies_filepath)

    async def logout(self, *, force=False):
        """
        End user session by calling :attr:`_logout`, which must be implemented by the subclass

        Set :attr:`is_logged_in` to `False` and emit `logged_out` :attr:`signal`, even if the logout
        request fails for some reason.

        If a :attr:`cookies_filepath` is provided, do nothing unless `force` is `True`.

        :raise errors.RequestError: on failure
        """
        if self.cookies_filepath and not force:
            _log.debug('%s: Not logging out because user session is stored: %r', self.name, self.cookies_filepath)
        else:
            async with self._session_lock:
                if self.is_logged_in:
                    caller = inspect.currentframe().f_back.f_code
                    _log.debug('%s: Logging out on behalf of %r', self.name, caller)
                    self.signal.emit('logging_out')
                    try:
                        await self._logout()
                    except errors.RequestError as e:
                        _log.debug('Logout failed: %r', e)
                        self.signal.emit('logout_failed', e)
                        raise
                    finally:
                        _log.debug('Logged out successfully')
                        self._is_logged_in = False
                        self.signal.emit('logged_out')

    @abc.abstractmethod
    async def _login(self):
        """
        Start user session

        :raise errors.RequestError: on any kind of failure
        """

    @abc.abstractmethod
    async def confirm_logged_in(self):
        """
        Check if we are logged in by doing a website request

        This method is called by :meth:`login` to make sure whatever :meth:`_login` did worked. It
        is also called by :meth:`still_logged_in`, which is used check if a stored user session is
        still working or if a new session must be started (e.g. because the cookie expired).

        :raise errors.RequestError: if we are not logged in
        """

    @abc.abstractmethod
    async def _logout(self):
        """
        End user session

        :raise errors.RequestError: on any kind of failure
        """

    @functools.cached_property
    def _session_lock(self):
        # Prevent multiple simultaneous login/logout attempts
        return asyncio.Lock()

    @property
    def is_logged_in(self):
        """
        Whether a user session is active

        This is a boolean flag that is set by :meth:`login` and :meth:`logout`.
        """
        return getattr(self, '_is_logged_in', False)

    @property
    def cookies_filepath(self):
        """
        User-defined file path that stores login session cookie(s) or `None` if no such file is
        specified

        Subclasses must include this in requests made to the tracker website if the tracker
        implementation uses traditional user sesssions via the website for submissions.

        If the tracker implementation uses an API, cookies should not be used in requests and this
        property can be ignored.
        """
        cookies_filepath = self.options.get('cookies_filepath')
        if cookies_filepath:
            return os.path.expanduser(cookies_filepath)

    def delete_cookies_filepath(self):
        """Delete :attr:`cookies_filepath` if it is provided and exists"""
        cookies_filepath = self.cookies_filepath
        if cookies_filepath:
            try:
                os.remove(cookies_filepath)
            except FileNotFoundError:
                pass
            else:
                _log.debug('%s: Cookies removed: %r', self.name, cookies_filepath)

    @abc.abstractmethod
    async def get_announce_url(self):
        """
        Get announce URL from :attr:`options` or tracker website

        .. warning:: It is highly recommended to call :meth:`login` first. Otherwise, getting the
                     URL from the tracker website should be impossible. :class:`RuntimeError` should
                     be raised if :attr:`is_logged_in` is `False`.

        :raise errors.RequestError: on any kind of failure
        """

    calculate_piece_size = None
    """
    :class:`staticmethod` that takes a torrent's content size and returns
    the corresponding piece size

    If this is `None`, the default implementation is used.
    """

    calculate_piece_size_min_max = None
    """
    :class:`staticmethod` that takes a torrent's content size and returns
    the corresponding allowed minimum and maximum piece sizes

    If this is `None`, the default implementation is used.
    """

    @abc.abstractmethod
    async def upload(self, tracker_jobs):
        """
        Upload torrent and other metadata from jobs

        :param TrackerJobsBase tracker_jobs: :attr:`TrackerJobs` instance
        """

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance with the following signals:

        ``warning``
            Emitted when :meth:`warn` is called. Registered callbacks get the provided `warning`
            argument.

        ``error``
            Emitted when :meth:`error` is called. Registered callbacks get the provided `error`
            argument.

        ``exception``
            Emitted when :meth:`exception` is called. Registered callbacks get the provided
            `exception` argument.

        ``logging_in``
            Emitted by :meth:`login` when login is attempted. Registered callbacks get no arguments.

        ``login_failed``
            Emitted by :meth:`login` when login failed. Registered callbacks get a
            :class:`~.RequestError` exception.

        ``logged_in``
            Emitted by :meth:`login` when login succeeded. Registered callbacks get no arguments.

        ``logging_out``
            Emitted by :meth:`logout` when logout is attempted. Registered callbacks get no
            arguments.

        ``logout_failed``
            Emitted by :meth:`logout` when logout failed. Registered callbacks get a
            :class:`~.RequestError` exception.

        ``logged_out``
            Emitted by :meth:`logout` when logout succeeded. Registered callbacks get no arguments.
        """
        return self._signal

    def warn(self, warning):
        """
        Emit ``warning`` signal (see :attr:`signal`)

        Emit a warning for any non-critical issue that the user can choose to
        ignore or fix.
        """
        self.signal.emit('warning', warning)

    def error(self, error):
        """
        Emit ``error`` signal (see :attr:`signal`)

        Emit an error for any critical but expected issue that can't be
        recovered from (e.g. I/O error).
        """
        self.signal.emit('error', error)

    def exception(self, exception):
        """
        Emit ``exception`` signal (see :attr:`signal`)

        Emit an exception for any critical and unexpected issue that should be
        reported as a bug.
        """
        self.signal.emit('exception', exception)



"""
Execute external commands
"""

import os
import pty
import selectors

from .. import errors
from ..utils import LazyModule

subprocess = LazyModule(module='subprocess', namespace=globals())


class Process:
    """
    Convenience wrapper around :class:`subprocess.Popen`

    The main benefit of it is that you can easily iterate over lines on stdout or stderr.
    """

    def __init__(self, popen):
        self._popen = popen

    def _iterlines(self, fh):
        # Tell readline() to return an empty string if there is nothing to read.
        os.set_blocking(fh.fileno(), False)

        selector = selectors.DefaultSelector()
        selector.register(fh.fileno(), selectors.EVENT_READ)

        while True:
            line = fh.readline()
            yield line
            if not line:
                is_running = self.is_running
                if not is_running:
                    # Process terminated and we consumed all buffered lines.
                    return
                else:
                    # This is essnentially `time.sleep(0.1)`, but the call is interrupted as soon as
                    # there is anything to read on `fh`. It means we can sleep for a long time but
                    # still read output when the subprocess writes it.
                    for _ in selector.select(timeout=0.1):
                        pass

    @property
    def stdout(self):
        """
        Iterator that yields lines from standard output

        If there is nothing to read, an empty string is yielded after a short timeout to prevent
        your loop from being blocked.
        """
        yield from self._iterlines(self._popen.stdout)

    @property
    def stderr(self):
        """
        Iterator that yields lines from standard error or nothing if standard error is
        redirected to standard out.

        If there is nothing to read, an empty string is yielded after a short timeout to prevent
        your loop from being blocked.
        """
        if self._popen.stderr:
            yield from self._iterlines(self._popen.stderr)
        else:
            # Popen.stderr is None if the "stderr" argument wasn't `subprocess.PIPE`.
            # This means stderr is redirected to stdout.
            yield from ()

    @property
    def is_running(self):
        """Whether the process is still running"""
        return self._popen.poll() is None

    def wait(self):
        """Wait for process to finish"""
        self._popen.wait()

    def terminate(self):
        """
        Ask process to terminate (SIGTERM) and kill it (SIGKILL) if it doesn't do that after 1
        second

        This method does nothing if the process is not running.
        """
        if self.is_running:
            self._popen.terminate()
            try:
                self._popen.wait(timeout=1)
            except subprocess.TimeoutExpired:
                self._popen.kill()

    @property
    def exitcode(self):
        """
        Exit code or termination signal after process ended

        See :attr:`subprocess.Popen.returncode`
        """
        return self._popen.returncode


def run(argv, *, ignore_errors=False, join_stderr=False, return_exitcode=False, communicate=False):
    """
    Execute command in subprocess

    :param argv: Command to execute
    :type argv: list of str
    :param bool ignore_errors: Do not raise :class:`.ProcessError` if stderr is
        non-empty
    :param bool join_stderr: Redirect stderr to stdout
    :param bool return_exitcode: Return a 2-tuple of `(stdout, exitcode)` instead of only `stdout`
    :param bool communicate: Instead of the command's output, return a :class:`~.subproc.Process`
        instance

    :raise DependencyError: if the command fails to execute
    :raise ProcessError: if stdout is not empty and `ignore_errors` is `False`

    :return: Output from process
    :rtype: str
    """
    argv = tuple(str(arg) for arg in argv)

    # STDOUT and STDERR
    stdout_argument = subprocess.PIPE
    if join_stderr:
        stderr_argument = subprocess.STDOUT
    else:
        stderr_argument = subprocess.PIPE

    # Use a pseudo TTY instead of subprocess.PIPE because BDInfo refuses to run otherwise.
    _pty_stdin_master, pty_stdin_slave = pty.openpty()

    try:
        process = Process(subprocess.Popen(
            argv,
            shell=False,
            text=True,
            universal_newlines=True,
            stdout=stdout_argument,
            stderr=stderr_argument,
            stdin=pty_stdin_slave,
            # Line buffering
            bufsize=1,
        ))
    except OSError as e:
        raise errors.DependencyError(f'Missing dependency: {argv[0]}') from e
    else:
        if communicate:
            return process
        else:
            process.wait()
            stdout = ''.join(process.stdout)
            stderr = ''.join(process.stderr)

            if stderr and not ignore_errors:
                raise errors.ProcessError(stderr)
            elif return_exitcode:
                return stdout, process.exitcode
            else:
                return stdout



"""
Swiss Army knife
"""

import asyncio
import collections
import functools
import hashlib
import importlib
import inspect
import itertools
import os
import re
import types as _types

import logging  # isort:skip
_log = logging.getLogger(__name__)


def is_running_in_development_environment():
    """Whether we are running in a development environment or in production"""
    return bool(os.environ.get('VIRTUAL_ENV', None))


def os_family():
    """
    Return "windows" or "unix"
    """
    return 'windows' if os.name == 'nt' else 'unix'


# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
class LazyModule(_types.ModuleType):
    """
    Lazily import module to decrease execution time

    :param str module: Name of the module
    :param mapping namespace: Usually the return value of `globals()`
    :param str name: Name of the module in `namespace`; defaults to `module`
    """

    def __init__(self, module, namespace, name=None):
        self._module = module
        self._namespace = namespace
        self._name = name or module
        super().__init__(module)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._namespace[self._name] = module

        # Update this object's dict so that if someone keeps a reference to the
        # LazyLoader, lookups are efficient (__getattr__ is only called on
        # lookups that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


def submodules(package):
    """
    Return list of submodules and subpackages in `package`

    :param str package: Fully qualified name of parent package,
        e.g. "upsies.utils.imghosts"
    """
    # Get absolute path to parent directory of top-level package
    own_path = os.path.dirname(__file__)
    rel_path = __package__.replace('.', os.sep)
    assert own_path.endswith(rel_path), f'{own_path!r}.endswith({rel_path!r})'
    project_path = own_path[:-len(rel_path)]

    # Add relative path within project to given package
    package_path = os.path.join(project_path, package.replace('.', os.sep))

    # Find and import public submodules
    submods = []
    for name in os.listdir(package_path):
        if not name.startswith('_'):
            name = name.removesuffix('.py')
            if '.' not in name:
                submods.append(
                    importlib.import_module(name=f'.{name}', package=package)
                )
    return submods


def subclasses(basecls, modules):
    """
    Find subclasses in modules

    :param type basecls: Class that all returned classes are a subclass of
    :param modules: Modules to search
    :type modules: list of module objects
    """
    subclses = []
    for mod in modules:
        for _name, member in inspect.getmembers(mod):
            if (member is not basecls and
                isinstance(member, type) and
                issubclass(member, basecls)):
                subclses.append(member)
    return tuple(subclses)


def closest_number(n, ns, max=None, default=0):
    """
    Return the number from `ns` that is closest to `n`

    :param n: Given number
    :param ns: Sequence of allowed numbers
    :param max: Remove any item from `ns` that is larger than `max`
    :param default: Return value in case `ns` is empty
    """
    if max is not None:
        ns_ = tuple(n_ for n_ in ns if n_ <= max)
        if not ns_:
            raise ValueError(f'No number equal to or below {max}: {ns}')
    else:
        ns_ = ns
    return min(ns_, key=lambda x: abs(x - n), default=default)


class CaseInsensitiveString(str):
    """String that ignores case when compared or sorted"""

    def __hash__(self):
        return hash(self.casefold())

    def __eq__(self, other):
        if not isinstance(other, str):
            return NotImplemented
        else:
            return self.casefold() == other.casefold()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.casefold() < other.casefold()

    def __le__(self, other):
        return self.casefold() <= other.casefold()

    def __gt__(self, other):
        return self.casefold() > other.casefold()

    def __ge__(self, other):
        return self.casefold() >= other.casefold()


class MonitoredList(collections.abc.MutableSequence):
    """
    :class:`list` that calls `callback` after every change

    :param callback: Callable that gets the instance as a positional argument
    """

    def __init__(self, *args, callback, **kwargs):
        self._list = list(*args, **kwargs)
        self._callback = callback

    def __getitem__(self, index):
        return self._list[index]

    def __setitem__(self, index, value):
        self._list[index] = value
        self._callback(self)

    def __delitem__(self, index):
        del self._list[index]
        self._callback(self)

    def insert(self, index, value):
        self._list.insert(index, value)
        self._callback(self)

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        return self._list == other

    def __repr__(self):
        return f'{type(self).__name__}({self._list!r}, callback={self._callback!r})'


def is_sequence(obj):
    """Return whether `obj` is a sequence and not a string"""
    return (
        isinstance(obj, collections.abc.Sequence)
        and not isinstance(obj, str)
    )


def merge_dicts(a, b, path=()):
    """
    Merge nested dictionaries `a` and `b` into new dictionary with same
    structure
    """
    keys = itertools.chain(a, b)
    merged = {}
    for key in keys:
        if (isinstance(a.get(key), collections.abc.Mapping) and
            isinstance(b.get(key), collections.abc.Mapping)):
            merged[key] = merge_dicts(a[key], b[key], (*path, key))
        elif key in b:
            # Value from b takes precedence
            merged[key] = b[key]
        elif key in a:
            # Value from a is default
            merged[key] = a[key]
    return merged


def deduplicate(seq, key=None):
    """
    Return sequence `seq` with all duplicate items removed while maintaining
    the original order

    :param key: Callable that gets each item and returns a hashable identifier
        for that item
    """
    if key is None:
        def key(k):
            return k

    seen_keys = set()
    deduped = []
    for item in seq:
        k = key(item)
        if k not in seen_keys:
            seen_keys.add(k)
            deduped.append(item)
    return deduped


def as_groups(sequence, group_sizes, default=None):
    """
    Iterate over items from `sequence` in equally sized groups

    :params sequence: List of items to group
    :params group_sizes: Sequence of acceptable number of items in a group

        Find the group size with the lowest number of `default` items in
        the last group. That group size is then used for all groups.
    :param default: Value to pad last group with if
        ``len(sequence) % group_size != 0``

    Example:

    >>> sequence = range(1, 10)
    >>> for group in as_groups(sequence, [4, 5], default="_"):
    ...     print(group)
    (1, 2, 3, 4, 5)
    (6, 7, 8, 9, '_')
    >>> for group in as_groups(sequence, [3, 4], default="_"):
    ...     print(group)
    (1, 2, 3)
    (4, 5, 6)
    (7, 8, 9)
    """
    # Calculate group size that results in the least number of `default` values
    # in the final group
    gs_map = collections.defaultdict(list)
    for gs in group_sizes:
        # How many items from `sequence` are in the last group
        overhang = len(sequence) % gs
        # How many `default` values are in the last group
        default_count = 0 if overhang == 0 else gs - overhang
        gs_map[default_count].append(gs)

    lowest_default_count = sorted(gs_map)[0]
    group_size = max(gs_map[lowest_default_count])
    args = [iter(sequence)] * group_size
    yield from itertools.zip_longest(*args, fillvalue=default)


_unsupported_semantic_hash_types = (
    collections.abc.Iterator,
    collections.abc.Iterable,
    collections.abc.Generator,
)

def semantic_hash(obj):
    """
    Return SHA256 hash for `obj` that stays the same between Python interpreter
    sessions

    https://github.com/schollii/sandals/blob/master/json_sem_hash.py
    """
    def as_str(obj):
        if isinstance(obj, str):
            return obj

        elif isinstance(obj, collections.abc.Mapping):
            stringified = ((as_str(k), as_str(v)) for k, v in obj.items())
            return as_str(sorted(stringified))

        elif isinstance(obj, (collections.abc.Sequence, collections.abc.Set)):
            stringified = (as_str(item) for item in obj)
            return ''.join(sorted(stringified))

        elif isinstance(obj, _unsupported_semantic_hash_types):
            raise RuntimeError(f'Unsupported type: {type(obj)}: {obj!r}')

        else:
            return str(obj)

    return hashlib.sha256(bytes(as_str(obj), 'utf-8')).hexdigest()


def run_task(coro, callback):
    """
    Run awaitable in background task and return immediately

    This method should be used to call coroutine functions and other awaitables
    in a synchronous context.

    The returned task must be collected (e.g. in a :class:`list`) and awaited or
    cancelled eventually.

    :param coro: Any awaitable object
    :param callback: Callable that is called with the returned task when
        `coro` returns, is cancelled or raises any other exception

    :return: :class:`asyncio.Task` instance
    """
    if not callable(callback):
        raise ValueError(f'Not callable: {callable!r}')

    def handle_task_done(task):
        # Call callback no matter what
        try:
            task.result()
        except BaseException:
            callback(task)
        else:
            callback(task)

    task = asyncio.create_task(coro)
    task.add_done_callback(handle_task_done)
    return task


async def run_async(function, *args, **kwargs):
    """
    Run synchronous `function` asynchronously in a thread

    See :meth:`asyncio.BaseEventLoop.run_in_executor`.
    """
    loop = asyncio.get_running_loop()
    wrapped = functools.partial(function, *args, **kwargs)
    return await loop.run_in_executor(None, wrapped)


_NOTHING = object()

def blocking_memoize(coro_func):
    """
    Asynchronous memoization decorator that blocks concurrent calls with the
    same arguments

    The first call calls the decorated function while subsequent calls wait
    until the first call returns and the return value is cached. Subsequent
    calls then get the return value from the cache.

    Exceptions raised by `coro_func` are also cached and re-raised on subsequent
    calls.

    The decorated function provides a `clear_cache` method that removes any
    cached return values.
    """
    cache = {}
    lock = collections.defaultdict(asyncio.Lock)

    @functools.wraps(coro_func)
    async def wrapper(*args, **kwargs):
        cache_key = semantic_hash((str(coro_func), args, kwargs))
        async with lock[cache_key]:
            result = cache.get(cache_key, _NOTHING)

            if result is _NOTHING:
                try:
                    result = await coro_func(*args, **kwargs)
                except BaseException as e:
                    result = e
                cache[cache_key] = result

            if isinstance(result, BaseException):
                raise result
            else:
                return result

    def clear_cache():
        cache.clear()

    wrapper.clear_cache = clear_cache

    return wrapper


def flatten_nested_lists(thing):
    """
    Return flattened :class:`list`

    :param thing: Arbitrarily nested iterables

    If `thing` is not an iterable (and not a string), it is returned inside a
    list.
    """
    def flatten(x):
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, (str, bytes)):
            for item in x:
                yield from flatten(item)
        else:
            yield x

    return list(flatten(thing))


# We must import these here to prevent circular imports
from . import (  # noqa: E402 isort:skip
    argtypes,
    browser,
    btclient,
    configfiles,
    country,
    daemon,
    disc,
    fs,
    html,
    http,
    image,
    imghosts,
    mediainfo,
    predbs,
    release,
    signal,
    string,
    subproc,
    torrent,
    types,
    update,
    webdbs,
)



import functools
import os
import re

from ... import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

langcodes = utils.LazyModule(module='langcodes', namespace=globals())

NO_DEFAULT_VALUE = object()


def has_dual_audio(path, default=NO_DEFAULT_VALUE):
    """
    Return `True` if `path` contains multiple audio tracks with different languages and one of
    them is English, `False` otherwise

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        audio_tracks = utils.mediainfo.lookup(path, ('Audio',))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        languages = set()
        for track in audio_tracks:
            if 'commentary' not in track.get('Title', '').lower():
                language = track.get('Language', '')
                if language:
                    # `language` is of format "LANGUAGE" or "LANGUAGE-REGION", where each "LANGUAGE"
                    # and "REGION" are two-character codes. (Probably BCP47?)
                    languages.add(language.casefold()[:2])
        return len(languages) > 1


def has_commentary(path, default=NO_DEFAULT_VALUE):
    """
    Return `True` if `path` has an audio track with "Commentary" (case-insensitive) in its
    ``Title`` field, `False` otherwise

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        audio_tracks = utils.mediainfo.lookup(path, ('Audio',))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        for track in audio_tracks:
            if _is_commentary(track):
                return True
        return False


def _is_commentary(track):
    title = track.get('Title', '')
    return bool(re.search(r'\b(commentary|comments)\b', title, flags=re.IGNORECASE))


def get_audio_languages(path, default=NO_DEFAULT_VALUE, *, exclude_commentary=True):
    """
    Return sequence of two-letter (ISO 639-1) language codes from audio tracks

    If an audio track does not specify a language, use `default` if specified, otherwise ignore that
    audio track.

    :param str path: Path to release files

        For directories, the return value of :func:`find_main_video` is used.

        BDMV and VIDEO_TS releases are supported. For multi-disc releases (i.e. `path` contains multiple
        directories with "BDMV" or "VIDEO_TS" subdirectories), languages from all discs are accumulated.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :param exclude_commentary: Ignore any track with a ``"Title"`` field that contains the string
        "commentary" (case-insensitive)

        .. warning:: Commentary detection is not supported for BDMV and VIDEO_TS releases.

    :raise ContentError: if anything goes wrong
    """
    languages = []

    if utils.disc.is_disc(path, multidisc=True):
        # BDMV or VIDEO_TS
        languages.extend(get_audio_languages_from_discs(
            path,
            default=default,
            exclude_commentary=exclude_commentary,
        ))

    else:
        # Regular video file(s)
        languages.extend(
            get_audio_languages_from_mediainfo(
                path,
                default=default,
                exclude_commentary=exclude_commentary,
            )
        )

    return tuple(languages)


def get_audio_languages_from_mediainfo(path, default=NO_DEFAULT_VALUE, *, exclude_commentary=True):
    """
    Return sequence of BCP47 language codes from audio tracks

    This function supports regular video files (.mkv, .mp4, .avi, etc).

    See :func:`get_audio_languages`.
    """
    try:
        audio_tracks = utils.mediainfo.lookup(path, ('Audio',))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        languages = []
        for track in audio_tracks:
            if not exclude_commentary or not _is_commentary(track):
                language = track.get('Language', '')
                if language:
                    # `language` should be in format "LANGUAGE" or "LANGUAGE-REGION" or "LANGUAGE /
                    # LANGUAGE" (I don't know what that means), where each "LANGUAGE" and "REGION"
                    # are two-character codes (probably BCP47). But we don't know for sure, so we
                    # fall back to adding the language as is if it's an invalid code.
                    try:
                        langcode = langcodes.Language.get(language[:2])
                    except ValueError:
                        languages.append(language)
                    else:
                        languages.append(langcode.language)
                elif default is not NO_DEFAULT_VALUE:
                    languages.append(default)
        return tuple(languages)


@functools.cache
def get_audio_languages_from_discs(path, default=NO_DEFAULT_VALUE, *, exclude_commentary=True):
    """
    Return sequence of two-letter (ISO 639-1) language codes from audio tracks

    This function reads audio languages from .MPLS or .IFO playlists in a "VIDEO_TS" or "BDMV"
    subdirectory. `path` may be a multidisc release.

    See :func:`get_audio_languages`.
    """
    if not os.path.exists(path):
        if default is NO_DEFAULT_VALUE:
            raise errors.ContentError(f'No such file or directory: {path}')
        else:
            return default
    else:
        if utils.disc.is_bluray(path, multidisc=True):
            disc_module = utils.disc.bluray
        elif utils.disc.is_dvd(path, multidisc=True):
            disc_module = utils.disc.dvd
        else:
            _log.debug('Not a disc path: %r', path)
            return ()

        languages = []
        for disc_path in disc_module.get_disc_paths(path):
            main_playlists = tuple(
                playlist
                for playlist in disc_module.get_playlists(disc_path)
                if playlist.is_main
            )
            for playlist in main_playlists:
                languages.extend(get_audio_languages_from_mediainfo(
                    playlist.filepath,
                    default=default,
                    exclude_commentary=exclude_commentary,
                ))
            _log.debug('Audio languages from disc tree: %r', languages)
        return tuple(languages)


_audio_format_translations = (
    # (<format>, <<key>:<regex> dictionary>)
    # - All <regex>s must match each <key> to identify <format>.
    # - All identified <format>s are appended (e.g. "TrueHD Atmos").
    # - {<key>: None} means <key> must not exist.
    ('AAC', {'Format': re.compile(r'^AAC$')}),
    ('DD', {'Format': re.compile(r'^AC-3$')}),
    ('DDP', {'Format': re.compile(r'^E-AC-3$')}),
    ('TrueHD', {'Format': re.compile(r'MLP ?FBA')}),
    ('TrueHD', {'Format_Commercial_IfAny': re.compile(r'TrueHD')}),
    ('Atmos', {'Format_Commercial_IfAny': re.compile(r'Atmos')}),
    ('DTS', {'Format': re.compile(r'^DTS$'), 'Format_Commercial_IfAny': None}),
    ('DTS-ES', {'Format_Commercial_IfAny': re.compile(r'DTS-ES')}),
    ('DTS-HD', {'Format_Commercial_IfAny': re.compile(r'DTS-HD(?! Master Audio)')}),
    ('DTS-HD MA', {
        'Format_Commercial_IfAny': re.compile(r'DTS-HD Master Audio'),
        'Format_AdditionalFeatures': re.compile(r'XLL$'),
    }),
    ('DTS:X', {'Format_AdditionalFeatures': re.compile(r'XLL X')}),
    ('FLAC', {'Format': re.compile(r'FLAC')}),
    ('MP3', {'Format': re.compile(r'MPEG Audio')}),
    ('Vorbis', {'Format': re.compile(r'\bVorbis\b')}),
    ('Vorbis', {'Format': re.compile(r'\bOgg\b')}),
    ('Opus', {'Format': re.compile(r'\bOpus\b')}),
    ('PCM', {'Format': re.compile(r'PCM')}),
)

@functools.cache
def get_audio_format(path, default=NO_DEFAULT_VALUE):
    """
    Return audio format of default audio track (e.g. "AAC", "DDP Atmos") or empty string if
    `path` has no audio track

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        # Check if `path` has any audio tracks at all.
        if 'Audio' in utils.mediainfo.get_tracks(path):
            audio_track = utils.mediainfo.lookup(path, ('Audio', 'DEFAULT'))
        else:
            _log.debug('No audio track found: %r', path)
            return ''

    except errors.ContentError as e:
        # `path` does not exist.
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default

    else:

        def is_match(regexs, audio_track):
            for key, regex in regexs.items():
                if regex is None:
                    if key in audio_track:
                        # `key` must not exists but it does: No match.
                        return False
                elif key not in audio_track:
                    # `key` does not exist.
                    return False
                elif not regex.search(audio_track.get(key, '')):
                    # `key` exists, but its value does not match `regex`.
                    return False
            # All `regexs` match and no forbidden keys in `audio_track`.
            return True

        parts = []
        for fmt, regexs in _audio_format_translations:
            if fmt not in parts and is_match(regexs, audio_track):
                parts.append(fmt)

        if parts:
            return ' '.join(parts)

        elif default is NO_DEFAULT_VALUE:
            raise errors.ContentError('Unable to detect audio format')

        else:
            _log.debug('Failed to detect audio format, falling back to default: %s', default)
            return default


_audio_channels_translations = (
    ('1.0', re.compile(r'^1$')),
    ('2.0', re.compile(r'^2$')),
    ('2.0', re.compile(r'^3$')),
    ('2.0', re.compile(r'^4$')),
    ('2.0', re.compile(r'^5$')),
    ('5.1', re.compile(r'^6$')),
    ('5.1', re.compile(r'^7$')),
    ('7.1', re.compile(r'^8$')),
    ('7.1', re.compile(r'^9$')),
    ('7.1', re.compile(r'^10$')),
)

def get_audio_channels(path, default=NO_DEFAULT_VALUE):
    """
    Return audio channels of default audio track (e.g. "5.1") or empty_string
    (e.g. if `path` has no audio track)

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        audio_track = utils.mediainfo.lookup(path, ('Audio', 'DEFAULT'))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        audio_channels = ''
        channels = audio_track.get('Channels', '')
        for achan, regex in _audio_channels_translations:
            if regex.search(channels):
                audio_channels = achan
                break
        return audio_channels



"""
Share generated metadata
"""

from .. import errors, trackers
from . import JobBase, JobRunner

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SubmitJob(JobBase):
    """
    Submit torrent file and other metadata to tracker

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``submitting``
            Emitted when attempting to upload metadata. Registered callbacks get no arguments.

        ``submitted``
            Emitted when upload attempt ended. Registered callbacks get the torrent page URL of the
            submitted torrent as a positional argument or a :class:`~.RequestError` if the upload
            failed.
    """

    name = 'submit'
    label = 'Submit'

    # Don't cache output.
    cache_id = None

    def initialize(self, *, tracker_jobs):
        """
        Set internal state

        :param TrackerBase tracker: Return value of :func:`~.trackers.tracker`
        :param TrackerJobsBase tracker_jobs: Instance of :attr:`~.TrackerBase.TrackerJobs`
        """
        assert isinstance(tracker_jobs, trackers.TrackerJobsBase), f'Not a TrackerJobsBase: {tracker_jobs!r}'
        self._tracker_jobs = tracker_jobs
        self._tracker = tracker_jobs.tracker

        # Custom signals.
        self.signal.add('submitting')
        self.signal.add('submitted')

        # Pass through signals from TrackerBase subclass.
        self._tracker.signal.register('warning', self.warn)
        self._tracker.signal.register('error', self.error)
        self._tracker.signal.register('exception', self.exception)

    async def run(self):
        # Collect all the required metadata.
        await self._wait_for_jobs_before_upload()

        # Don't submit if self._tracker_jobs thinks that's a bad idea, e.g. if we only want the
        # release description.
        if self._tracker_jobs.submission_ok:
            await self._submit()

        await self._start_jobs_after_upload()

    async def _submit(self):
        _log.debug('Submitting')
        try:
            self.signal.emit('submitting')
            torrent_page_url = await self._tracker.upload(self._tracker_jobs)
            _log.debug('Torrent page url: %r', torrent_page_url)
            if torrent_page_url:
                self.add_output(torrent_page_url)
            self.signal.emit('submitted', torrent_page_url)
        except errors.RequestError as e:
            self.error(e)
        finally:
            _log.debug('Done submitting')

    async def _wait_for_jobs_before_upload(self):
        runner = JobRunner(self._tracker_jobs.jobs_before_upload, id='submit')
        await runner.wait()

    async def _start_jobs_after_upload(self):
        for job in self._tracker_jobs.jobs_after_upload:
            if job and not job.is_started and job.is_enabled:
                _log.debug('Starting job after upload: %r', job)
                job.start()

    @property
    def _enabled_jobs_before_upload(self):
        """
        Sequence of jobs to do before submission

        This is the same as :attr:`.TrackerJobsBase.jobs_before_upload` but with all `None` values
        and disabled jobs filtered out.
        """
        return tuple(
            job
            for job in self._tracker_jobs.jobs_before_upload
            if job and job.is_enabled
        )

    @property
    def hidden(self):
        """
        Hide this job if :attr:`~.TrackerJobsBase.submission_ok` is falsy

        If :attr:`~.TrackerJobsBase.submission_ok` is falsy, that usually means we are only
        generating some metadata that is not supposed to be submitted.

        It also should mean this job is not displayed until all
        :attr:`~.TrackerJobsBase.jobs_before_upload` have finished successfully.
        """
        return not self._tracker_jobs.submission_ok

    @property
    def _main_job(self):
        # Return an objects that provides `output` and `exit_code`.
        #
        # If submission is prevented by `TrackerJobsBase.submission_ok`, this is the last job of
        # `TrackerJobsBase.jobs_before_upload` if all those jobs are finished.
        #
        # Otherwise, return a :class:`super` instance, meaning we are using the parent class
        # (`JobBase`) to get `output` and `exit_code`.
        if not self._tracker_jobs.submission_ok:
            enabled_jobs = self._enabled_jobs_before_upload
            # Because jobs can enable/disable each other, we can't know the main job until all jobs
            # are either finished or disabled.
            if enabled_jobs and all(job.is_finished for job in enabled_jobs):
                return enabled_jobs[-1]
        return super()

    @property
    def output(self):
        """Output from :attr:`_main_job`"""
        return self._main_job.output

    @property
    def exit_code(self):
        """
        Exit code from last enabled job in :attr:`~.TrackerJobsBase.jobs_before_upload` if not
        :attr:`~.TrackerJobsBase.submission_ok`

        :attr:`.JobBase.exit_code` otherwise.
        """
        return self._main_job.exit_code



import functools

from prompt_toolkit.layout.containers import DynamicContainer

from .... import utils
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PlaylistsJobWidget(JobWidgetBase):

    is_interactive = True

    def setup(self):
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            extend_width=True,
        )
        self._queued_list_dialogs = []
        self._current_list_dialog = None

        self.job.signal.register('discs_available', self.handle_discs_available)
        self.job.signal.register('playlists_available', self.handle_playlists_available)
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

    def handle_discs_available(self, discpaths):
        self.add_list_dialog(
            options=tuple(
                (
                    # Human-readable value.
                    utils.fs.basename(path),
                    # Computer-usable value.
                    path,
                )
                for path in discpaths
            ),
            # Mark all discs as autodetected.
            autodetected_indexes=range(len(discpaths)),
            callback=self.job.discs_selected,
        )

    def handle_playlists_available(self, discpath, playlists):
        options = []
        autodetected_indexes = []
        for index, playlist in enumerate(playlists):
            options.append(
                (
                    # Human-readable value.
                    '   '.join((
                        str(utils.fs.basename(playlist.filepath)).rjust(10),
                        str(playlist.duration),
                        str(playlist.size).rjust(9),
                    )),
                    # Computer-usable value.
                    playlist,
                )
            )
            if playlist.is_main:
                autodetected_indexes.append(index)

        self.add_list_dialog(
            question=discpath,
            options=options,
            autodetected_indexes=autodetected_indexes,
            callback=lambda playlists: self.job.playlists_selected(discpath, playlists),
        )

    def add_list_dialog(self, *, options, callback, question=None, autodetected_indexes=()):
        if self.job.select_multiple:
            list_dialog = widgets.CheckList(
                question=question,
                options=options,
                on_accepted=lambda options: self.handle_checklist_dialog_accepted(callback, options),
                autodetected_indexes=autodetected_indexes,
                focused=(autodetected_indexes[0] if autodetected_indexes else None),
            )
        else:
            list_dialog = widgets.RadioList(
                question=question,
                options=options,
                on_accepted=lambda option: self.handle_radiolist_dialog_accepted(callback, option),
                autodetected_index=autodetected_indexes[0],
                focused=(autodetected_indexes[0] if autodetected_indexes else None),
            )
        self._queued_list_dialogs.append(list_dialog)
        self.maybe_show_next_list_dialog()

    def maybe_show_next_list_dialog(self):
        if not self._current_list_dialog and self._queued_list_dialogs:
            self._current_list_dialog = self._queued_list_dialogs.pop(0)
            self.invalidate()

    def handle_checklist_dialog_accepted(self, callback, options):
        # The first item in each 2-tuple option is the human-readable display value, the second item
        # is what we actually use internally.
        values = tuple(option[1] for option in options)
        callback(values)
        self._current_list_dialog = None
        self.maybe_show_next_list_dialog()

    def handle_radiolist_dialog_accepted(self, callback, option):
        # The first item in `option` the human-readable display value, the second item is what we
        # actually use internally.
        values = (option[1],)
        callback(values)
        self._current_list_dialog = None
        self.maybe_show_next_list_dialog()

    @functools.cached_property
    def runtime_widget(self):
        return DynamicContainer(self.get_runtime_widgets)

    def get_runtime_widgets(self):
        if self._current_list_dialog:
            self._activity_indicator.disable()
            return self._current_list_dialog
        else:
            self._activity_indicator.enable()
            return self._activity_indicator



"""
Hardcoded PTP metadata from upload.php
"""

import enum
import re

tags = (
    'action',
    'adventure',
    'animation',
    'arthouse',
    'asian',
    'biography',
    'camp',
    'comedy',
    'crime',
    'cult',
    'documentary',
    'drama',
    'experimental',
    'exploitation',
    'family',
    'fantasy',
    'film.noir',
    'history',
    'horror',
    'martial.arts',
    'musical',
    'mystery',
    'performance',
    'philosophy',
    'politics',
    'romance',
    'sci.fi',
    'short',
    'silent',
    'sport',
    'thriller',
    'video.art',
    'war',
    'western',
)


types = {
    'Feature Film': re.compile(r'^(?i:(?:feature|).?film|movie)$'),
    'Short Film': re.compile(r'^(?i:short.?(?:film|))$'),
    'Miniseries': re.compile(r'^(?i:(?:mini.?|)series|season|tv)$'),
    'Stand-up Comedy': re.compile(r'^(?i:stand.?up.?comedy|stand.?up|comedy)$'),
    'Live Performance': re.compile(r'(?i:live.?performance|live|performance)$'),
    'Movie Collection': re.compile(r'(?i:movie.?collection|collection)$'),
}


sources = {
    'Blu-ray': re.compile(r'(?i:blu-?ray|bd(?:25|50|66|100))'),  # (UHD) BluRay|BD(25|50|66|100) (Remux)
    'DVD': re.compile(r'^(?i:dvd)'),  # DVD(Rip|5|9|...)
    'WEB': re.compile(r'^(?i:web)'),  # WEB(-DL|Rip)
    'HD-DVD': re.compile(r'^(?i:hd-?dvd)$'),
    'HDTV': re.compile(r'^(?i:hd-?tv)$'),
    'TV': re.compile(r'^(?i:(?:sd-?|)tv)'),  # TV(Rip|...)
    'VHS': re.compile(r'^(?i:vhs)'),  # VHS(Rip|...)
}


# NOTE: The order of the dictionaries and their elements are copied from
#       upload.php. They should be submitted in the same order as listed here.
editions = {
    'collection.criterion': 'The Criterion Collection',
    'collection.masters': 'Masters of Cinema',
    'collection.warner': 'Warner Archive Collection',

    'edition.dc': "Director's Cut",
    'edition.extended': 'Extended Edition',
    'edition.rifftrax': 'Rifftrax',
    'edition.theatrical': 'Theatrical Cut',
    'edition.uncut': 'Uncut',
    'edition.unrated': 'Unrated',

    'feature.remux': 'Remux',
    'feature.2in1': '2in1',
    'feature.2disc': '2-Disc Set',
    'feature.3d_anaglyph': '3D Anaglyph',
    'feature.3d_full_sbs': '3D Full SBS',
    'feature.3d_half_ou': '3D Half OU',
    'feature.3D_half_sbs': '3D Half SBS',
    'feature.4krestoration': '4K Restoration',
    'feature.10bit': '10-bit',
    'feature.extras': 'Extras',
    'feature.4kremaster': '4K Remaster',
    'feature.2d3d_edition': '2D/3D Edition',
    'feature.dtsx': 'DTS:X',
    'feature.dolby_atmos': 'Dolby Atmos',
    'feature.dolby_vision': 'Dolby Vision',
    'feature.hdr10': 'HDR10',
    'feature.hdr10+': 'HDR10+',
    'feature.dual_audio': 'Dual Audio',
    'feature.english_dub': 'English Dub',
    'feature.commentary': 'With Commentary',
}


subtitles = {
    # Special codes
    'No Subtitles': '44',
    'en (forced)': '50',

    # TODO: No idea what intertitles are exactly and how to detect them.
    # 'English Intertitles': '51',

    # Regular languages
    'ar': '22',     # Arabic
    'bg': '29',     # Bulgarian
    'zh': '14',     # Chinese
    'hr': '23',     # Croatian
    'cs': '30',     # Czech
    'da': '10',     # Danish
    'nl': '9',      # Dutch
    'en': '3',      # English
    'et': '38',     # Estonian
    'fi': '15',     # Finnish
    'fr': '5',      # French
    'de': '6',      # German
    'el': '26',     # Greek
    'he': '40',     # Hebrew
    'hi': '41',     # Hindi
    'hu': '24',     # Hungarian
    'is': '28',     # Icelandic
    'id': '47',     # Indonesian
    'it': '16',     # Italian
    'ja': '8',      # Japanese
    'ko': '19',     # Korean
    'lv': '37',     # Latvian
    'lt': '39',     # Lithuanian
    'no': '12',     # Norwegian
    'fa': '52',     # Persian
    'pl': '17',     # Polish
    'pt': '21',     # Portuguese
    'pt-BR': '49',  # Brazilian Port.
    'ro': '13',     # Romanian
    'ru': '7',      # Russian
    'sr': '31',     # Serbian
    'sk': '42',     # Slovak
    'sl': '43',     # Slovenian
    'es': '4',      # Spanish
    'sv': '11',     # Swedish
    'th': '20',     # Thai
    'tr': '18',     # Turkish
    'uk': '34',     # Ukrainian
    'vi': '25',     # Vietnamese
}


class _PrettyEnum(enum.Enum):
    @classmethod
    def from_string(cls, string):
        """
        Convert human-readable string back to enum

        >>> TrumpableReason.from_string(
        ...     str(TrumpableReason.HARDCODED_SUBTITLES)
        ... )
        <TrumpableReason.HARDCODED_SUBTITLES: 4>

        :raise AttributeError: if `string` is not known
        """
        name = string.replace(' ', '_').upper()
        return getattr(cls, name)

    def __str__(self):
        return ' '.join(
            word.capitalize()
            for word in self.name.split('_')
        )


class TrumpableReason(_PrettyEnum):
    """
    Reason why a release is trumpable

    An instance's :attr:`value` is the value expected by the API. An instance's
    string representation should be human-readable and pretty.
    """

    NO_ENGLISH_SUBTITLES = 14
    HARDCODED_SUBTITLES = 4


class ArtistImportance(_PrettyEnum):
    """
    Purpose of a person in a movie

    An instance's :attr:`value` is the value expected by the API. An instance's
    string representation should be human-readable and pretty.
    """

    ACTOR = 5
    DIRECTOR = 1
    WRITER = 2
    PRODUCER = 3
    COMPOSER = 4
    CINEMATOGRAPHER = 6



import functools

from . import JobWidgetBase

# This job is hidden so we only need a dummy widget.

class SetJobWidget(JobWidgetBase):

    is_interactive = False

    def setup(self):
        pass

    @functools.cached_property
    def runtime_widget(self):
        return None



from ... import errors
from ..base import rules


class BhdBannedGroup(rules.BannedGroup):

    banned_groups = {
        '4K4U',
        'AOC',
        'C4K',
        'd3g',
        'EASports',
        'FGT',  # Unless no other encode is available.
        'MeGusta',
        'MezRips',
        'nikt0',
        'ProRes',
        'RARBG',
        'ReaLHD',
        'SasukeducK',
        'Sicario',
        'TEKNO3D',  # They have requested their torrents are not shared off site.
        'Telly',
        'tigole',
        'TOMMY',
        'WKS',
        'x0r',
        'YIFY',
    }

    async def _check_custom(self):
        # No iFT remuxes.
        if (
                self.is_group('iFT')
                and 'Remux' in self.release_name.source
        ):
            raise errors.BannedGroup('iFT', additional_info='No remuxes from iFT')

        # No EVO encodes. WEB-DLs are fine.
        if (
                self.is_group('EVO')
                and 'WEB' not in self.release_name.source
        ):
            raise errors.BannedGroup('EVO', additional_info='No encodes, only WEB-DL')



"""
Concrete :class:`~.TrackerBase` subclass for ANT
"""

import math
import urllib

from ... import errors, utils
from ..base import TrackerBase
from . import rules
from .config import AntTrackerConfig
from .jobs import AntTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class AntTracker(TrackerBase):
    name = 'ant'
    label = 'ANT'
    torrent_source_field = 'ANT'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. API Key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to USERNAME -> Edit -> Access Settings\n'
        '       and scroll down to "API keys".\n'
        '   {howto.current_section}.2 In the "Create a new Key" row, tick the "Upload" box.\n'
        '   {howto.current_section}.3 Click on "Save profile".\n'
        '   {howto.current_section}.4 Scroll down to "API keys" again and copy the new API_KEY.\n'
        '   {howto.current_section}.5 $ upsies set trackers.{tracker.name}.apikey API_KEY\n'
        '\n'
        '{howto.next_section}. Announce URL\n'
        '\n'
        '   {howto.current_section}.1 On the website, click on "Upload" and copy the ANNOUNCE_URL.\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.announce_url ANNOUNCE_URL\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = AntTrackerConfig
    TrackerJobs = AntTrackerJobs
    rules = rules

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _api_url(self):
        return urllib.parse.urljoin(self._base_url, '/api.php')

    @property
    def apikey(self):
        apikey = self.options.get('apikey')
        if apikey:
            return apikey
        else:
            raise errors.RequestError('No API key configured')

    async def _login(self, *, tfa_otp=None):
        pass

    async def confirm_logged_in(self):
        pass

    async def _logout(self):
        pass

    async def get_announce_url(self):
        announce_url = self.options.get('announce_url')
        if announce_url:
            return announce_url
        else:
            raise errors.AnnounceUrlNotSetError(tracker=self)

    async def upload(self, tracker_jobs):
        post_data = tracker_jobs.post_data

        _log.debug('POSTing data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        post_files = tracker_jobs.post_files
        _log.debug('POSTing files: %r', post_files)

        json = await self._request(
            method='POST',
            url=self._api_url,
            cache=False,
            data=post_data,
            files=post_files,
        )

        if json.get('status') == 'success':
            return tracker_jobs.torrent_filepath
        elif json.get('error'):
            raise errors.RequestError(f'Upload failed: {json["error"]}')
        else:
            raise RuntimeError(f'Unexpected response: {json!r}')

    async def _request(self, method, *args, **kwargs):
        try:
            # `method` is "GET" or "POST"
            response = await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                **kwargs,
            )
        except errors.RequestError as e:
            _log.debug(f'Request failed: {e!r}')
            _log.debug(f'url={e.url!r}')
            _log.debug(f'text={e.text!r}')
            _log.debug(f'headers={e.headers!r}')
            _log.debug(f'status_code={e.status_code!r}')
            # The error message in the HTTP response is JSON. Try to parse that
            # to get the actual error message. If that fails, raise the
            # RequestError as is.
            json = e.json(default=None)
            if json:
                return json
            else:
                raise e
        else:
            return response.json()

    @staticmethod
    def calculate_piece_size(bytes):
        # Recommended piece count is 1000.
        exponent = math.ceil(math.log2(bytes / 1000))

        # Piece size range: 1 MiB - 64 MiB
        exponent = max(20, min(26, exponent))
        return int(math.pow(2, exponent))

    @staticmethod
    def calculate_piece_size_min_max(bytes):
        # Maximum torrent file size is 100 KiB. We reserve a maximum of 60 KiB
        # for piece hashes, leaving 40 KiB for other metadata. Each piece hash
        # is 20 bytes long.
        max_piece_count = (60 * 1024) / 20
        min_exponent = math.ceil(math.log2(bytes / max_piece_count))

        # Absolute minimum piece size: 1 MiB
        min_exponent = max(20, min_exponent)

        # For large torrents, the minimum piece size can end up larger then the
        # maximum. Here we limit the minimum piece size to 8 MiB or less.
        min_exponent = min(23, min_exponent)

        return (
            math.pow(2, min_exponent),
            64 * 2**20,  # 64 MiB
        )



import functools

from .... import errors
from . import JobWidgetBase


class RulesJobWidget(JobWidgetBase):

    def setup(self):
        self.job.signal.register('error', self._handle_error)

    def _handle_error(self, error):
        if isinstance(error, errors.RuleBroken):
            self.job.error('You can override rule checks with --ignore-rules.')

    @functools.cached_property
    def runtime_widget(self):
        return None



"""
Custom TUI widgets
"""

import asyncio
import collections
import itertools
import textwrap

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import ANSI, to_formatted_text
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit, VSplit, Window, WindowAlign
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension

import logging  # isort:skip
_log = logging.getLogger(__name__)


hspacer = Window(FormattedTextControl(' '), dont_extend_width=True, dont_extend_height=True)
vspacer = Window(FormattedTextControl(''), dont_extend_width=True, dont_extend_height=True)


class TextField:
    """Single line of non-editable text"""

    def __init__(self, text='', *, width=None, height=1, extend_width=True, style=''):
        self._text = text
        self._width = width
        self._height = height
        self._activity_indicator = ActivityIndicator(callback=self.set_text)
        self.container = Window(
            content=FormattedTextControl(lambda: self.text),
            width=width,
            height=height,
            dont_extend_height=True,
            dont_extend_width=not extend_width,
            wrap_lines=True,
            style=style,
        )

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        self.set_text(text)

    def set_text(self, text):
        if isinstance(self._width, int) and self._width > 1:
            width = self._width
            if isinstance(self._height, int):
                height = self._height
            else:
                height = None

            if height == 1:
                # Remove newlines. Don't do word wrapping to avoid replacing
                # long words with "…" with lots of available free space.
                self._text = text.replace('\n', ' ')
            else:
                # prompt_toolkit has no support for word wrapping.
                paragraphs = text.split('\n')
                self._text = '\n\n'.join(textwrap.fill(
                    paragraph,
                    width=width,
                    max_lines=height,
                    placeholder='…',
                ) for paragraph in paragraphs)

        else:
            self._text = text

        get_app().invalidate()

    @property
    def is_loading(self):
        """Whether an activity indicator is displayed"""
        return self._activity_indicator.active

    @is_loading.setter
    def is_loading(self, is_loading):
        self._activity_indicator.active = is_loading

    def __pt_container__(self):
        return self.container


class InputField:
    """Single line of user-editable text"""

    def __init__(self, text='', *, width=None, extend_width=True, read_only=False,
                 on_accepted=None, on_changed=None, style=''):
        self._activity_indicator = ActivityIndicator()
        self.read_only = read_only
        self.on_accepted = on_accepted
        self.buffer = Buffer(
            multiline=False,
            accept_handler=self._accept_handler,
            on_text_changed=on_changed,
            read_only=Condition(lambda: self.read_only),
        )
        self.container = VSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._activity_indicator.active),
                    content=VSplit([
                        self._activity_indicator,
                        hspacer,
                    ]),
                ),
                Window(
                    content=BufferControl(self.buffer),
                    always_hide_cursor=Condition(lambda: self.read_only),
                    width=width,
                    dont_extend_height=True,
                    dont_extend_width=not extend_width,
                ),
            ],
            style=style,
        )
        if text:
            self.set_text(text, ignore_callback=True)

    def _accept_handler(self, buffer):
        if self.on_accepted:
            self.on_accepted(buffer)
        # Do not clear input field on enter
        return True

    @property
    def text(self):
        return self.buffer.text

    @text.setter
    def text(self, text):
        self.set_text(text)

    def set_text(self, text, *, ignore_callback=False, preserve_cursor_position=False):
        if preserve_cursor_position:
            curpos = self.buffer.cursor_position

        if ignore_callback:
            handlers = self.buffer.on_text_changed._handlers
            self.buffer.on_text_changed._handlers = []

        self.buffer.set_document(Document(text), bypass_readonly=True)

        if ignore_callback:
            self.buffer.on_text_changed._handlers = handlers

        if preserve_cursor_position:
            self.buffer.cursor_position = curpos

    @property
    def read_only(self):
        return self._read_only

    @read_only.setter
    def read_only(self, read_only):
        self._read_only = bool(read_only)

    @property
    def is_loading(self):
        """Whether an activity indicator is displayed"""
        return self._activity_indicator.active

    @is_loading.setter
    def is_loading(self, is_loading):
        self._activity_indicator.active = is_loading

    def __pt_container__(self):
        return self.container


class RadioList:
    """
    List of options from which the user can select one

    :param options: Sequence of options the user can make
    :param question: Text that is displayed above the `options` or any falsy
        value
    :param focused: Focused index or option or `None` to focus the first option
    :param on_accepted: Callback that gets the user choice

    Each option item must be a :class:`str` or a sequence with one or more
    items. The first item is presented to the user. `on_accepted` gets the
    complete option item.
    """

    def __init__(self, options=(), question=None, focused=None, autodetected_index=None, on_accepted=None):
        self.options = options
        self.question = question
        self.on_accepted = on_accepted
        self.autodetected_index = autodetected_index

        if isinstance(focused, int):
            # It seems prompt-toolkit can't handle negative indexes.
            if focused < 0:
                self.focused_index = len(options) + focused
            else:
                self.focused_index = focused
        elif focused:
            self.focused_index = options.index(focused)
        else:
            self.focused_index = 0

        kb = KeyBindings()

        @kb.add('up')
        @kb.add('c-p')
        @kb.add('k')
        @kb.add('s-tab')
        def _(_event):
            self.focused_index = max(0, self.focused_index - 1)

        @kb.add('down')
        @kb.add('c-n')
        @kb.add('j')
        @kb.add('tab')
        def _(_event):
            self.focused_index = min(len(self.options) - 1, self.focused_index + 1)

        @kb.add('enter')
        @kb.add('c-j')
        def _(_event):
            if self.on_accepted is not None:
                try:
                    self.on_accepted(self.options[self.focused_index])
                except IndexError as e:
                    raise RuntimeError(f'No option at index {self.focused_index}: {self.options}') from e

        choices = Window(
            content=FormattedTextControl(
                self._get_text_fragments,
                key_bindings=kb,
                focusable=True,
            ),
            dont_extend_height=True,
            always_hide_cursor=True,
        )

        self.container = HSplit(
            children=[
                choices,
            ],
            style='class:dialog.choice',
        )

        if self.question:
            self.container.children.insert(0, Window(FormattedTextControl([
                ('bold', self.question),
            ])))

    @property
    def focused_option(self):
        """Currently focused item in :attr:`options`"""
        return self.options[self.focused_index]

    @focused_option.setter
    def focused_option(self, option):
        if option not in self.options:
            raise ValueError(f'No such option: {option!r}')
        else:
            self.focused_index = self.options.index(option)

    def _get_text_fragments(self):
        fragments = []
        if not self.options:
            return fragments

        def option_as_string(option):
            if isinstance(option, str):
                return option
            elif isinstance(option, collections.abc.Sequence):
                return str(option[0])
            else:
                raise RuntimeError(f'Invalid option value: {option!r}')

        width = max(len(option_as_string(option)) for option in self.options)
        for i, option in enumerate(self.options):
            option_string = option_as_string(option)
            if i == self.focused_index:
                style = 'class:dialog.choice.focused'
                fragments.append(('[SetCursorPosition]', ''))
                fragments.append((style, ' \N{BLACK CIRCLE}'))
            else:
                style = 'class:dialog.choice'
                fragments.append((style, '  '))
            fragments.append((style, ' '))
            fragments.extend(to_formatted_text(option_string.ljust(width) + ' ', style=style))
            if i == self.autodetected_index:
                fragments.append(('class:dialog.choice', ' (autodetected)'))
            fragments.append(('', '\n'))

        fragments.pop()  # Remove last newline

        return fragments

    def __pt_container__(self):
        return self.container


class CheckList:
    """
    List of options from which the user can select multiple

    :param options: Sequence of options the user can make
    :param question: Text that is displayed above the `options` or any falsy
        value
    :param focused: Focused index or option or `None` to focus the first option
    :param on_accepted: Callback that gets the ticked options

    Each option item must be a :class:`str` or a sequence with one or more
    items. The first item is presented to the user. `on_accepted` gets the
    complete option item.
    """

    def __init__(self, options=(), question=None, focused=None, autodetected_indexes=(), on_accepted=None):
        self.options = tuple(options)
        self.question = str(question) if question else ''
        self.on_accepted = on_accepted

        if isinstance(focused, int):
            # It seems prompt-toolkit can't handle negative indexes.
            if focused < 0:
                self.focused_index = len(options) + focused
            else:
                self.focused_index = focused
        elif focused:
            self.focused_index = options.index(focused)
        else:
            self.focused_index = 0

        self.marked_indexes = set(autodetected_indexes)
        self.autodetected_indexes = set(autodetected_indexes)

        kb = KeyBindings()

        @kb.add('up')
        @kb.add('c-p')
        @kb.add('k')
        @kb.add('s-tab')
        def _(_event):
            self.focused_index = max(0, self.focused_index - 1)

        @kb.add('down')
        @kb.add('c-n')
        @kb.add('j')
        @kb.add('tab')
        def _(_event):
            self.focused_index = min(len(self.options) - 1, self.focused_index + 1)

        @kb.add(' ')
        def _(_event):
            if self.focused_index in self.marked_indexes:
                self.marked_indexes.remove(self.focused_index)
            else:
                self.marked_indexes.add(self.focused_index)
            self.focused_index = min(len(self.options) - 1, self.focused_index + 1)

        @kb.add('enter')
        @kb.add('c-j')
        def _(_event):
            if self.on_accepted is not None:
                self.on_accepted(self.marked_options)

        self.container = HSplit(
            children=[
                Window(
                    content=FormattedTextControl(
                        self._get_text_fragments,
                        key_bindings=kb,
                        focusable=True,
                    ),
                    dont_extend_height=True,
                    dont_extend_width=True,
                    always_hide_cursor=True,
                ),
            ],
            style='class:dialog.choice',
        )

        if self.question:
            self.container.children.insert(0, Window(FormattedTextControl([
                ('bold', self.question),
            ])))

    @property
    def marked_options(self):
        """Currently marked items in :attr:`options`"""
        return tuple(
            option
            for index, option in enumerate(self.options)
            if index in self.marked_indexes
        )

    def _get_text_fragments(self):
        fragments = []
        if not self.options:
            return fragments

        def option_as_string(option):
            if isinstance(option, str):
                return option
            elif isinstance(option, collections.abc.Sequence):
                return str(option[0])
            else:
                raise RuntimeError(f'Invalid option value: {option!r}')

        width = max(len(option_as_string(option)) for option in self.options)
        for i, option in enumerate(self.options):
            option_string = option_as_string(option)
            if i == self.focused_index:
                style = 'class:dialog.choice.focused'
                fragments.append(('[SetCursorPosition]', ''))
            else:
                style = 'class:dialog.choice'

            if i in self.marked_indexes:
                fragments.append((style, ' \N{FISHEYE} '))
            else:
                fragments.append((style, ' \N{WHITE CIRCLE} '))

            fragments.append((style, ' '))
            fragments.extend(to_formatted_text(option_string.ljust(width) + ' ', style=style))
            if i in self.autodetected_indexes:
                fragments.append(('class:dialog.choice', ' (autodetected)'))

            fragments.append(('', '\n'))

        fragments.pop()  # Remove last newline
        return fragments

    def __pt_container__(self):
        return self.container


class HLabel:
    _groups = collections.defaultdict(list)

    def __init__(self, text, content, group=None, style=''):
        self.label = Window(
            content=FormattedTextControl(text),
            dont_extend_width=True,
            dont_extend_height=True,
            align=WindowAlign.RIGHT,
        )
        self.container = VSplit(
            [
                # Apply `style` to label text and spacer for background color.
                VSplit([self.label, hspacer], style=style),
                content,
            ],
        )

        self._group = group
        if group is not None:
            self._groups[group].append(self)
            texts = [label.label.content.text for label in self._groups[group]]
            max_width = max(len(text) for text in texts)
            for label in self._groups[group]:
                label.label.width = max_width + 1

    def __pt_container__(self):
        return self.container


class VLabel:
    def __init__(self, text, content, style=''):
        self.container = HSplit([
            Window(
                FormattedTextControl(text=text),
                dont_extend_width=False,
                dont_extend_height=True,
                style=style,
            ),
            content,
        ])

    def __pt_container__(self):
        return self.container


class ProgressBar:
    def __init__(self, text='', width=None):
        self.percent = 0
        self.container = VSplit(
            children=[
                Window(
                    style='class:info.progressbar.progress',
                    width=lambda: Dimension(weight=int(max(0, self.percent))),
                    height=1,
                ),
                Window(
                    style='class:info.progressbar',
                    width=lambda: Dimension(weight=int(100 - min(100, self.percent))),
                    height=1,
                ),
            ],
            width=(
                Dimension(min=10, max=60, preferred=30)
                if width is None else
                Dimension.exact(width)
            ),
        )

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, value):
        self._percent = float(value)

    def __pt_container__(self):
        return self.container


class ActivityIndicator:
    """
    Activity indicator that cycles through a sequence of strings

    :param callable callback: Callable that receives a single string from
        `states` as a positional argument
    :param states: Sequence of strings that are cycled through and passed to
        `callback` individually
    :param float interval: Delay between calls to `callback`
    :param str format: See :attr:`format`
    :param str style: prompt-toolkit style
    :param bool extend_width: Whether to maximize horizontally
    """

    def __init__(self, *, callback=None, states=('⠷', '⠯', '⠟', '⠻', '⠽', '⠾'),
                 interval=0.3, format='{indicator}', style='', extend_width=False):
        self._iterator = itertools.cycle(states)
        self._interval = float(interval)
        self._callback = callback or None
        self._format = format
        self._text = self._format.format(indicator=next(self._iterator))
        self._iterate_task = None

        self.container = Window(
            content=FormattedTextControl(lambda: self.text),
            dont_extend_height=True,
            dont_extend_width=not extend_width,
            style=style,
        )

    @property
    def active(self):
        """
        Whether :attr:`text` is continuously updated and `callback` called

        .. note:: This updates :attr:`text` in an asynchronous task. If you
            try to set this attribute outside of an asynchronous context, you
            get ``RuntimeError: no running event loop``.
        """
        return bool(self._iterate_task)

    @active.setter
    def active(self, active):
        if active:
            if not self._iterate_task:
                self._iterate_task = asyncio.get_running_loop().create_task(self._iterate())
        elif self._iterate_task:
            self._iterate_task.cancel()
            self._iterate_task = None

    def enable(self):
        """Calling this method sets :attr:`active` to `True`"""
        self.active = True

    def disable(self):
        """Calling this method sets :attr:`active` to `False`"""
        self.active = False

    @property
    def format(self):
        """
        Format string

        ``{indicator}`` is replaced with the current state in `states`.
        """
        return self._format

    @format.setter
    def format(self, format):
        self._format = format

    @property
    def text(self):
        """Formatted :attr:`format` string"""
        return self._text

    async def _iterate(self):
        while True:
            self._text = self._format.format(indicator=next(self._iterator))
            get_app().invalidate()

            if self._callback:
                self._callback(self._text)

            await asyncio.sleep(self._interval)

    def __pt_container__(self):
        return self.container


class Image:
    def __init__(self, bytes=b'', width=None, height=None):
        self._activity_indicator = ActivityIndicator()
        self._width = width
        self._height = height
        self.container = VSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self.is_loading),
                    content=self._activity_indicator,
                ),
                ConditionalContainer(
                    filter=Condition(lambda: not self.is_loading),
                    content=Window(
                        content=FormattedTextControl(lambda: self._image),
                        dont_extend_height=True,
                        dont_extend_width=True,
                    ),
                ),
            ],
        )

        self.is_loading = False
        self.set_image_bytes(bytes)

    def set_image_bytes(self, bytes):
        if bytes:
            import io

            import PIL
            import term_image.image

            # Prevent DecompressionBombWarning: Image size (... pixels) exceeds
            # limit of ... pixels, could be decompression bomb DOS attack.
            # This happens when a poster is very large.
            PIL.Image.MAX_IMAGE_PIXELS = None

            escape_codes = str(
                # FIXME: This hack doesn't work with KittyImage
                # objects. Probably because the ANSI class doesn't know the
                # kitty graphics protocol and interprets the escape sequences as
                # normal text. To fix this, we always use BlockImage, even on
                # more capable terminals.
                term_image.image.BlockImage(
                    PIL.Image.open(io.BytesIO(bytes)),
                    width=self._width,
                    height=self._height,
                )
            )
        else:
            escape_codes = ''
        self._image = ANSI(escape_codes)

    # Add `text` property like InputField, TextField, etc. This is useful for
    # WebDbSearchJobWidget, which handles a lot of widgets. Avoiding a special
    # case for the poster widget makes things a lot simpler.
    @property
    def text(self):
        return self._image

    @text.setter
    def text(self, bytes):
        self.set_image_bytes(bytes)

    @property
    def is_loading(self):
        return self._activity_indicator.active

    @is_loading.setter
    def is_loading(self, value):
        self._activity_indicator.active = bool(value)

    def __pt_container__(self):
        return self.container



import re

from ... import errors, utils
from ..base import rules


class MtvBannedGroup(rules.BannedGroup):
    banned_groups = {
        '3LTON',
        '[Oj]',
        'aXXo',
        'BDP',
        'BRrip',
        'CM8',
        'CMCT',
        'CrEwSaDe',
        'DeadFish',
        'DNL',
        'ELiTE',
        'FaNGDiNG0',
        'FRDS',
        'FUM',
        'h65',
        'HD2DVD',
        'HDTime',
        'ION10',
        'iPlanet',
        'JIVE',
        'KiNGDOM',
        'Leffe',
        'LOAD',
        'mHD',
        'mRS',
        'mSD',
        'NhaNc3',
        'nHD',
        'nikt0',
        'nSD',
        'PandaRG',
        'PRODJi',
        'QxR',
        'RARBG',
        'RDN',
        'SANTi',
        'STUTTERSHIT',
        'TERMiNAL',
        'TM',
        'WAF',
        'x0r',
        'XS',
        'YIFY',
        'ZKBL',
        'ZmN',
    }

    async def _check_custom(self):
        # No EVO encodes. WEB-DLs are fine.
        if (
                self.is_group('EVO')
                and 'WEB' not in self.release_name.source
        ):
            raise errors.BannedGroup('EVO', additional_info='No encodes, only WEB-DL (Rule 2.2.1)')

        # No XviD encodes from ViSiON.
        if (
                self.is_group('ViSiON')
                and 'XviD' in self.release_name.video_format
        ):
            raise errors.BannedGroup('ViSiON', additional_info='No XviD encodes')


class MtvBannedContainerFormat(rules.TrackerRuleBase):
    _banned_file_extensions = {
        '3ivx',
        'asf',
        'f4v',
        'flv',
        'mov',
        'ogg',
        'qt',
        'rm',
        'rmv',
        'wmv',
    }

    async def _check(self):
        for file in utils.fs.file_list(self.release_name.path):
            file_extension = utils.fs.file_extension(file)
            if file_extension.casefold() in self._banned_file_extensions:
                raise errors.RuleBroken(f'Banned container format: {file_extension}: {file}')


class MtvHevcUhdOnly(rules.TrackerRuleBase):
    """HEVC (H.265) is only allowed for UHD (2160p), except for some anime groups"""

    _exempt_groups = {
        'AC',
        'Aergia',
        'ARC',
        'Arid',
        'Baws',
        'Chihiro',
        'Commie',
        'Crow',
        'CsS',
        'Dae',
        'Datte13',
        'Drag',
        'FLE',
        'GJM',
        'GJM-Kaleido',
        'hchcsen',
        'iKaos',
        'JySzE',
        'Kaleido-Subs',
        'Legion',
        'LostYears',
        'MTBB',
        'Netaro',
        'Noyr',
        'Okay-Subs',
        'OZR',
        'Reza',
        'sam',
        'Spirale',
        'Thighs',
        'TTGA',
        'UDF',
        'UQW',
        'Vanilla',
        'WSE',
    }

    async def _check(self):
        is_h265 = bool(re.search(r'(?:(?:H\.|x)265|HEVC)', self.release_name.video_format))
        resolution = utils.mediainfo.video.get_resolution_int(self.release_name.path)

        # H.265 is always OK for 2160p.
        if is_h265 and resolution < 2160:
            release_name_group = self.release_name.group.casefold()
            group_is_exempt = any(
                group.casefold() == release_name_group
                for group in self._exempt_groups
            )

            # For all non-exempt groups, H.265 is only allowed for UHD.
            if not group_is_exempt:
                raise errors.RuleBroken('HEVC is only allowed for UHD')

            # For some groups, H.265 is allowed, but only for HD.
            elif resolution < 720:
                raise errors.RuleBroken('HEVC is only allowed for HD')



from ... import errors
from ..base import rules


class UhdHdOnly(rules.HdOnly):
    pass


class UhdNoEnglishDubs(rules.TrackerRuleBase):
    """Check if Dual Audio release"""

    async def _check(self):
        if self.release_name.has_dual_audio:
            raise errors.RuleBroken('No English dubs except for animated content')



"""
Generate all required metadata and upload to tracker
"""

import functools
import os

from .... import __project_name__, constants, jobs, trackers, utils
from . import base


class submit(base.CommandBase):
    """Generate all required metadata and upload to TRACKER"""

    names = ('submit',)

    argument_definitions = {}

    subcommand_name = 'TRACKER'
    subcommands = {
        tracker.name: {
            'description': (
                f'Generate all required metadata and upload it to {tracker.label}.\n'
                '\n'
                f'For step-by-step instructions run this command:\n'
                '\n'
                f'    $ {__project_name__} submit {tracker.name} --howto-setup\n'
            ),
            'cli': {
                # Default arguments for all tackers
                ('--howto-setup',): {
                    'action': base.PrintText(text_getter=tracker.generate_setup_howto),
                    'nargs': 0,
                    'help': 'Show detailed instructions on how to do your first upload',
                },
                'CONTENT': {
                    'type': utils.argtypes.content,
                    'help': 'Path to release content',
                },
                ('--is-scene',): {
                    'type': utils.argtypes.bool_or_none,
                    'default': None,
                    'help': ('Whether this is a scene release (usually autodetected)\n'
                             'Valid values: '
                             + ', '.join(
                                 f'{true}/{false}'
                                 for true, false in zip(utils.types.Bool.truthy, utils.types.Bool.falsy)
                             )),
                },
                ('--exclude-files', '--ef'): {
                    'nargs': '+',
                    'action': 'extend',
                    'metavar': 'PATTERN',
                    'help': ('Glob pattern to exclude from torrent '
                             '(matched case-insensitively against path in torrent)'),
                    'default': [],
                },
                ('--exclude-files-regex', '--efr'): {
                    'nargs': '+',
                    'action': 'extend',
                    'metavar': 'PATTERN',
                    'help': ('Regular expression to exclude from torrent '
                             '(matched case-sensitively against path in torrent)'),
                    'type': utils.argtypes.regex,
                    'default': [],
                },
                ('--reuse-torrent', '-t'): {
                    'nargs': '+',
                    'metavar': 'TORRENT',
                    'help': ('Use hashed pieces from TORRENT instead of generating '
                             'them again or getting them from '
                             f'{utils.fs.tildify_path(constants.GENERIC_TORRENTS_DIRPATH)}\n'
                             'TORRENT may also be a directory, which is searched recursively '
                             'for a matching *.torrent file.\n'
                             "NOTE: This option is ignored if TORRENT doesn't match properly."),
                    'type': utils.argtypes.existing_path,
                    'default': (),
                },
                ('--add-to', '-a'): {
                    'type': utils.argtypes.client,
                    'metavar': 'CLIENT',
                    'help': ('Case-insensitive BitTorrent client name\n'
                             'Supported clients: ' + ', '.join(utils.btclient.client_names())),
                },
                ('--copy-to', '-c'): {
                    'metavar': 'PATH',
                    'help': 'Copy the created torrent to PATH (file or directory)',
                },
                ('--ignore-rules', '--ir'): {
                    'action': 'store_true',
                    'help': 'Allow submission if it is against tracker rules',
                },
                # Custom arguments defined by tracker for this command
                **tracker.TrackerConfig.argument_definitions.get('submit', {}),
            },
        }
        for tracker in trackers.trackers()
    }

    @functools.cached_property
    def jobs(self):
        return (
            *self.tracker_jobs.jobs_before_upload,
            self.main_job,
            *self.tracker_jobs.jobs_after_upload,
        )

    @functools.cached_property
    def main_job(self):
        return jobs.submit.SubmitJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            tracker_jobs=self.tracker_jobs,
        )

    @functools.cached_property
    def tracker_name(self):
        """Lower-case abbreviation of tracker name"""
        return self.args.subcommand.lower()

    @functools.cached_property
    def tracker_options(self):
        """
        :attr:`tracker_name` section in trackers configuration file combined with
        CLI arguments where CLI arguments take precedence unless their value is
        `None`
        """
        return self.get_options('trackers', self.tracker_name)

    @functools.cached_property
    def tracker(self):
        """
        :class:`~.trackers.base.TrackerBase` instance from one of the submodules of
        :mod:`.trackers`
        """
        return trackers.tracker(
            name=self.tracker_name,
            options=self.tracker_options,
        )

    @functools.cached_property
    def tracker_jobs(self):
        """
        :class:`~.trackers.base.TrackerJobsBase` instance from one of the submodules
        of :mod:`.trackers`
        """
        return self.tracker.TrackerJobs(
            tracker=self.tracker,
            options=self.tracker_options,
            content_path=self.args.CONTENT,
            reuse_torrent_path=(
                tuple(self.args.reuse_torrent)
                + tuple(self.config['config']['torrent-create']['reuse_torrent_paths'])
            ),
            screenshots_optimization=self.config['config']['screenshots']['optimize'],
            show_poster=self.config['config']['id']['show_poster'],
            image_hosts=self._get_image_hosts(),
            btclient=self._get_btclient(),
            torrent_destination=self._get_torrent_destination(),
            exclude_files=(
                tuple(self.config['trackers'][self.tracker.name]['exclude'])
                + tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
            common_job_args={
                'home_directory': self.home_directory,
                'cache_directory': self.cache_directory,
                'ignore_cache': self.args.ignore_cache,
            },
        )

    def _get_image_hosts(self):
        return tuple(
            self._get_image_host(name)
            for name in self.tracker_options.get('image_host', ())
        )

    def _get_image_host(self, name):
        # Get global image host options from imghosts.ini
        name = str(name).lower()
        options = self.config['imghosts'][name].copy()

        # Apply tracker-specific options that are common for all image hosts
        # (e.g. thumbnail size)
        options.update(self.tracker.TrackerJobs.image_host_config.get('common', {}))

        # Apply tracker-specific options for the used image host
        options.update(self.tracker.TrackerJobs.image_host_config.get(name, {}))

        return utils.imghosts.imghost(
            name=name,
            options=options,
            cache_directory=self.home_directory,
        )

    def _get_btclient_name(self):
        return (
            getattr(self.args, 'add_to', None)
            or self.tracker_options.get('add_to', None)
            or None
        )

    def _get_btclient(self):
        client_name = self._get_btclient_name()
        if client_name:
            options = self.get_options('clients', client_name)
            return utils.btclient.BtClient(
                name=client_name,
                url=options['url'],
                username=options['username'],
                password=options['password'],
                # See torrent_create.download_path for some reasoning.
                download_path=options['translate_path'].translate(
                    os.path.abspath(utils.fs.dirname(self.args.CONTENT))
                ),
                check_after_add=options['check_after_add'],
                category=options.get('category', ''),
            )

    def _get_torrent_destination(self):
        return (
            getattr(self.args, 'copy_to', None)
            or self.tracker_options.get('copy_to', None)
            or None
        )



"""
HTML parsing
"""

import re
import warnings

from . import LazyModule

bs4 = LazyModule(module='bs4', namespace=globals())


def parse(string):
    """
    Return :class:`~.bs4.BeautifulSoup` instance

    :param string: HTML document

    :raise ContentError: if `string` is invalid HTML
    """
    if isinstance(string, bs4.element.Tag):
        return string
    else:
        # Disable bs4 warnings, e.g. "MarkupResemblesLocatorWarning: The input
        # looks more like a filename than markup. You may want to open this file
        # and pass the filehandle into Beautiful Soup."
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=UserWarning)
            return bs4.BeautifulSoup(string, features='html.parser')


def dump(html, filepath):
    """
    Write `html` to `filepath` for debugging

    :param html: String or :class:`~.bs4.BeautifulSoup` instance
    """
    with open(filepath, 'w') as f:
        if isinstance(html, bs4.BeautifulSoup):
            f.write(html.prettify())
        else:
            f.write(parse(str(html)).prettify())


def get(soup, *attributes):
    """
    Get `attributes` from `soup`

    These two calls are equivalent if all attributes exist:

    >>> soup.table.tr.td
    "td value"
    >>> html.get(soup, "table", "tr", "td")
    "td value"

    But if any attribute is `None` (which is what :class:`~.bs4.BeautifulSoup`
    returns for unknown tags), you get `None` instead of forcing you to catch an
    :class:`AttributeError`:

    >>> soup.table.no_such_attribute.td
    AttributeError: 'NoneType' object has no attribute 'td'
    >>> html.get(soup, "table", "no_such_attribute", "td")
    None
    """
    for attr in attributes:
        soup = getattr(soup, attr, None)
        if soup is None:
            return None
    return soup


def as_text(html):
    """Strip HTML tags from string and return text without markup"""
    html = re.sub(r'<br\s*/?>', '\n', str(html))
    text = parse(html).get_text()
    text = re.sub(r'(\s)\s+', r'\1', text, flags=re.MULTILINE).strip()
    return text


def purge_tags(html):
    """
    Return `html` with <script> and <style> tags removed

    :param str html: HTML string
    """
    def is_javascript(tag):
        # Match <script> tags that don't have type="application/ld+json".
        if tag.name == 'script':
            return tag.get('type') != 'application/ld+json'

        # Match <style> (CSS) tags.
        elif tag.name == 'style':
            return True

        return False

    soup = parse(html)
    for script_tag in soup.find_all(is_javascript):
        script_tag.decompose()
    return str(soup)



"""
Tracker rules
"""

import asyncio

from .. import errors
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RulesJob(JobBase):
    """
    Check if a particulare release does not violate tracker rules

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``checking``
            Emitted when a rule is checked. Registered callbacks get the :class:`~.TrackerRuleBase`
            subclass as a positional argument.

        ``checked``
            Emitted when a rule was checked. Registered callbacks get the :class:`~.TrackerRuleBase`
            subclass and a :class:`~.RuleBroken` instance or `None` as positional arguments.
    """

    name = 'rules'
    label = 'Rules'

    # TODO: Can we cache this?
    cache_id = None

    @property
    def tracker(self):
        """Instance of a :class:`~.TrackerBase` subclass"""
        return self._tracker_jobs.tracker

    @property
    def tracker_jobs(self):
        """Instance of a :class:`~.TrackerJobsBase` subclass"""
        return self._tracker_jobs

    @property
    def release_name(self):
        """:class:`~.ReleaseName` instance"""
        return self._tracker_jobs.release_name

    # This job produces no output, only errors or warnings.
    no_output_is_ok = True

    # Errors and warnings are displayed even for hidden jobs.
    hidden = True

    def initialize(self, *, tracker_jobs, only_warn=False):
        """
        Set internal state

        :param tracker_jobs: Instance of a :class:`~.TrackerJobsBase` subclass

        :param bool only_warn: Exceptions from broken rules are passed to :meth:`~.JobBase.warn`
            instead of :meth:`~.JobBase.error` (i.e. allow submissions with broken rules)
        """
        self._tracker_jobs = tracker_jobs
        self._only_warn = only_warn
        self._rules_checked = {}
        self.signal.add('checking')
        self.signal.add('checked')

    async def run(self):
        await asyncio.gather(*(
            self.check_rule(TrackerRule)
            for TrackerRule in self.tracker.rules
        ))

    async def check_rule(self, TrackerRule):
        rule = TrackerRule(self.tracker_jobs)
        self.signal.emit('checking', rule)
        try:
            await rule.check()
        except errors.RuleBroken as e:
            _log.debug('%s: Rule broken: %r', TrackerRule.__name__, e)
            if self._only_warn:
                self.warn(e)
            else:
                self.error(e)
        else:
            _log.debug('%s: Rule obeyed', TrackerRule.__name__)
        self.signal.emit('checked', rule)



"""
Wrapper for ``bdinfo`` command
"""

import functools
import os
import re
import tempfile

from .. import errors, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BdinfoJob(base.JobBase):
    """
    Provide (partial) BDInfo report

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``progress``
            Emitted in short intervals after :meth:`playlists_selected` was called at least once
            until all BDInfo reports are available. Registered callbacks get a
            :class:`BdinfoProgress` instance.

        ``bdinfo_report``
            Emitted when a BDInfo report is available. Registered callbacks get a
            :class:`~.BdinfoReport` instance.
    """

    name = 'bdinfo'
    label = 'BDInfo'

    # Caching is done by the daemon process.
    cache_id = None

    _DEFAULT_FORMAT = '{BDINFO}'

    def initialize(self, *, summary=None, format=_DEFAULT_FORMAT):
        """
        Set internal state

        :param summary: Only provide a shorter summary as output

            One of ``full``, ``quick`` or `None`.

            Note that this only affects the :attr:`output`. :attr:`bdinfo_reports`,
            :attr:`full_summaries` and :attr:`quick_summaries` are always available.

        :param format: String that contains the placeholder ``"{BDINFO}"``, which is replaced by the
            actual BDInfo report
        """
        self._summary = summary
        self._format = str(format)
        self._bdinfo_reports = []
        self._reports_by_file = {}
        self._bdinfo_process = None

        self.signal.add('bdinfo_progress')
        self.signal.add('bdinfo_report')
        self.signal.register('bdinfo_report', self._store_bdinfo_report)
        self.signal.register('finished', self._hide_job)

    def _hide_job(self, job):
        # Do not display the BDInfo report as output. It can be very long.
        self.hidden = True

    async def run(self):
        self._bdinfo_process = utils.daemon.DaemonProcess(
            target=_bdinfo_process,
            kwargs={
                'cache_directory': self.cache_directory,
                'ignore_cache': self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._bdinfo_process.start()

        # Receive selected playlists from PlaylistsJob and pass them on to _bdinfo_process().
        async for discpath, playlists in self.receive_all('playlists', 'playlists_selected', only_posargs=True):
            _log.debug('Sending playlists to _bdinfo_process(): %r', playlists)
            self._bdinfo_process.send(utils.daemon.MsgType.info, {'playlists_selected': (discpath, playlists)})
        _log.debug('All playlists received')

        # Tell _bdinfo_process() that there won't be any more playlists.
        self._bdinfo_process.send(utils.daemon.MsgType.info, {'playlists_selected': _ALL_PLAYLIST_SELECTIONS_MADE})
        await self._bdinfo_process.join()
        self._add_bdinfo_reports()

    def _add_bdinfo_reports(self):
        if not self._summary:
            reports = self.bdinfo_reports
        elif self._summary == 'full':
            reports = self.full_summaries
        elif self._summary == 'quick':
            reports = self.quick_summaries
        else:
            raise RuntimeError(f'Unexpected summary value: {self._summary!r}')

        for report in reports:
            self.add_output(self._format.format(BDINFO=report))

    def terminate(self, reason=None):
        if self._bdinfo_process:
            self._bdinfo_process.stop()
        super().terminate(reason=reason)

    @base.unless_job_is_finished
    def _handle_info(self, info):
        if 'bdinfo_progress' in info:
            self.signal.emit('bdinfo_progress', info['bdinfo_progress'])

        elif 'bdinfo_report' in info:
            self.signal.emit('bdinfo_report', info['bdinfo_report'])

        else:
            raise RuntimeError(f'Unexpected info: {info!r}')

    @base.unless_job_is_finished
    def _handle_error(self, error):
        if isinstance(error, errors.DependencyError):
            self.error(error)
        else:
            self.exception(error)

    def _store_bdinfo_report(self, bdinfo_report):
        self._bdinfo_reports.append(bdinfo_report)
        self._reports_by_file[bdinfo_report.playlist.filepath] = bdinfo_report

    @property
    def bdinfo_reports(self):
        """Sequence of :class:`~.BdinfoReport` instances"""
        return tuple(self._bdinfo_reports)

    @property
    def reports_by_file(self):
        """Map playlist filepaths (.mpls) to BDInfo reports gathered so far"""
        return self._reports_by_file.copy()

    @property
    def full_summaries(self):
        """Sequence of :attr:`~.BdinfoReport.full_summary` values from :attr:`bdinfo_reports`"""
        return tuple(report.full_summary for report in self.bdinfo_reports)

    @property
    def quick_summaries(self):
        """Sequence of :attr:`~.BdinfoReport.quick_summary` values from :attr:`bdinfo_reports`"""
        return tuple(report.quick_summary for report in self.bdinfo_reports)


class BdinfoProgress(dict):
    """
    Simple :class:`dict` subclass with the following keys:

    - ``playlist`` (:class:`~.disc.Playlist`)
    - ``percent`` (:class:`int`)
    - ``time_elapsed`` (:class:`~.types.Timestamp`)
    - ``time_remaining`` (:class:`~.types.Timestamp`)

    Keys are also conveniently available as attributes.
    """

    def __init__(self, *, playlist, percent=0, time_elapsed=0, time_remaining=0):
        self['playlist'] = playlist
        self['percent'] = int(percent)
        self['time_elapsed'] = utils.types.Timestamp(time_elapsed)
        self['time_remaining'] = utils.types.Timestamp(time_remaining)

    def __getattr__(self, name):
        return self[name]


class BdinfoReport(str):
    """
    BDInfo report as special string with the additional properties :attr:`full_summary`,
    :attr:`quick_summary` and :attr:`playlist`
    """

    def __new__(cls, report, playlist):
        self = super().__new__(cls, report)
        self._playlist = playlist
        return self

    def __getnewargs__(self):
        # Make instances picklable by providing positional arguments.
        return (str(self), self.playlist)

    @functools.cached_property
    def full_summary(self):
        """Full summary of the BDInfo report"""
        regex = re.compile(r'^.*(DISC\s+INFO:$.*?)^FILES:$.*', flags=re.DOTALL | re.MULTILINE)
        match = regex.search(self)
        if match:
            return match.group(1).strip()

    @functools.cached_property
    def quick_summary(self):
        """Quick summary of the BDInfo report"""
        regex = re.compile(r'^.*^QUICK\s+SUMMARY:(.*)$', flags=re.DOTALL | re.MULTILINE)
        match = regex.search(self)
        if match:
            return match.group(1).strip()

    @functools.cached_property
    def playlist(self):
        """Source of the BDInfo report as a :class:`~.Playlist` instance"""
        return self._playlist

    def __repr__(self):
        return f'<{type(self).__name__} {self.playlist!r}>'


# NOTE: This cannot be the typical `object()` constant because its identity differs between the main
# process and _bdinfo_process() and `... is _ALL_PLAYLIST_SELECTIONS_MADE` doesn't work as expected.
_ALL_PLAYLIST_SELECTIONS_MADE = '_ALL_PLAYLIST_SELECTIONS_MADE'


def _bdinfo_process(output_queue, input_queue, *, cache_directory, ignore_cache):
    # This process reads user-selected playlists from the main process and runs `bdinfo` on them.
    while True:
        playlists_selected = utils.daemon.read_input_queue_key(input_queue, 'playlists_selected')

        if playlists_selected == _ALL_PLAYLIST_SELECTIONS_MADE:
            # No more values will be sent on `input_queue`.
            break
        else:
            _discpath, playlists = playlists_selected
            for playlist in playlists:
                bdinfo = _get_bdinfo(
                    output_queue, input_queue,
                    playlist=playlist,
                    cache_directory=cache_directory,
                    ignore_cache=ignore_cache,
                )
                bdinfo_report = BdinfoReport(bdinfo, playlist)
                output_queue.put((utils.daemon.MsgType.info, {'bdinfo_report': bdinfo_report}))
                utils.daemon.maybe_terminate(input_queue)


def _get_bdinfo(output_queue, input_queue, *, playlist, cache_directory, ignore_cache):
    # Make sure this is a BDMV path or we get weird exceptions.
    if not utils.disc.is_bluray(playlist.discpath):
        output_queue.put((utils.daemon.MsgType.error, errors.ContentError(f'Not a Blu-ray disc path: {playlist.discpath}')))

    else:
        cache_filepath = _get_cache_filepath(playlist, cache_directory)
        if (
                not ignore_cache
                and (bdinfo := _get_bdinfo_from_cache(cache_filepath))
        ):
            return bdinfo
        else:
            bdinfo = _get_bdinfo_from_bdinfo(output_queue, input_queue, playlist)
            _write_bdinfo_to_cache(bdinfo, cache_filepath)
            return bdinfo


def _get_bdinfo_from_bdinfo(output_queue, input_queue, playlist):
    # Write BDInfo report to temporary directory that will be deleted automatically.
    with tempfile.TemporaryDirectory() as bdinfo_directory:
        # Because BDInfo may be running inside a docker container with a different user ID, we
        # can't restrict permissions or the report cannot be written. This should not be an
        # issue because the BDInfo report isn't private information.
        os.chmod(bdinfo_directory, 0o777)

        for progress in _generate_bdinfo(
                playlist=playlist,
                bdinfo_directory=bdinfo_directory,
        ):
            utils.daemon.maybe_terminate(input_queue)
            output_queue.put((utils.daemon.MsgType.info, {'bdinfo_progress': progress}))

        return _find_bdinfo_in_directory(bdinfo_directory)


def _get_cache_filepath(playlist, cache_directory):
    return os.path.join(
        cache_directory,
        f'{playlist.label}.{int(playlist.size)}.bdinfo',
    )


def _get_bdinfo_from_cache(cache_filepath):
    try:
        with open(cache_filepath, 'r') as f:
            return f.read()
    except OSError as e:
        msg = e.strerror if e.strerror else str(e)
        _log.debug(f'Failed to read cache file {cache_filepath}: {msg}')


def _write_bdinfo_to_cache(bdinfo, cache_filepath):
    try:
        with open(cache_filepath, 'w') as f:
            return f.write(bdinfo)
    except OSError as e:
        msg = e.strerror if e.strerror else str(e)
        _log.debug(f'Failed to write cache file {cache_filepath}: {msg}')


def _generate_bdinfo(*, playlist, bdinfo_directory):
    # Initial zero progress report while `bdinfo` executable is starting up, which can take a few
    # seconds. This allows the UI to switch to a "report is being generated" display instead of just
    # doing nothing.
    yield BdinfoProgress(playlist=playlist)

    # Execute bdinfo with `communicate=True` so we can yield progress reports while it is running.
    argv = (
        'bdinfo',
        '--mpls=' + playlist.filename,
        playlist.discpath,
        bdinfo_directory,
    )
    process = utils.subproc.run(argv=argv, communicate=True)

    # Monitor bdinfo scanning progress.
    regex = re.compile(r'\D*?\b(?P<percent>\d+)(?:\.\d+|)\s?%.*?(?P<elapsed>\d+:\d+:\d+).*?(?P<remaining>\d+:\d+:\d+)')
    time_elapsed = utils.types.Timestamp(0)
    try:
        for line in process.stdout:
            match = regex.search(line)
            if match:
                time_elapsed = match.group('elapsed')
                yield BdinfoProgress(
                    playlist=playlist,
                    percent=match.group('percent'),
                    time_elapsed=time_elapsed,
                    time_remaining=match.group('remaining'),
                )

        stderr = '\n'.join(process.stderr)
        if stderr:
            raise RuntimeError(f'Command failed: {argv}:\n{stderr}')

        # Final progress report so the UI can reliably tell if a report is fully generated. This may
        # be redundant, but we can't rely on upstream bdinfo executable to always print "100%".
        yield BdinfoProgress(
            playlist=playlist,
            percent=100,
            time_elapsed=time_elapsed,
            time_remaining=0,
        )

    finally:
        # Always make sure the bdinfo process is terminated. Otherwise, it may continue to run even
        # after our main Python process terminates.
        process.terminate()


def _find_bdinfo_in_directory(bdinfo_directory):
    # Find report file. It should be the only file in `bdinfo_directory`.
    files = os.listdir(bdinfo_directory)
    if len(files) < 1:
        raise RuntimeError(f'No BDInfo report found in temporary BDInfo report directory: {bdinfo_directory}')
    elif len(files) > 1:
        raise RuntimeError(f'Unexpected files in temporary BDInfo report directory: {bdinfo_directory}: {sorted(files)}')
    else:
        bdinfo_file_path = os.path.join(bdinfo_directory, files[0])
        with open(bdinfo_file_path, 'r') as f:
            return f.read()



"""
Generic user dialogs

Prompts take arbitrary keyword arguments that are used by the user interface
to create the dialog.

When the user makes a choice, the user interface must pass it on to the prompt
object via :meth:`~.Prompt.set_result`. The prompt object then takes care of
passing the result to any callbacks.
"""

import asyncio

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Prompt:
    """
    Base class of all prompts

    Every prompt takes a sequence of callbacks. Any other keyword arguments
    are specified by the subclass. The user interface can access them as
    :attr:`parameters` to create UI widgets.

    Prompt instances are awaitable. The `await` will return when the user
    provided input. The return value of the `await` call is the same as the
    :attr:`result` property. Prompts can be awaited multiple times at any
    time.
    """

    def __init__(self, *, callbacks=(), **parameters):
        self._callbacks = list(callbacks)
        self._parameters = parameters
        self._result_arrived = asyncio.Event()
        self._result = None

    @property
    def parameters(self):
        """Keyword arguments from instantiation (except for `callbacks`)"""
        return self._parameters

    async def wait(self):
        """Block until :meth:`set_result` is called"""
        await self._result_arrived.wait()

    def __await__(self):
        async def get_result():
            await self.wait()
            return self.result
        return get_result().__await__()

    def on_result(self, callback):
        """Schedule `callback` to be called when :meth:`set_result` is called"""
        self._callbacks.append(callback)

    def set_result(self, result):
        """
        Take the result from the user dialog and make it available to
        callbacks and via the :attr:`result` property
        """
        _log.debug('Setting result: %r', result)
        self._result = result

        # Call callbacks
        for callback in self._callbacks:
            _log.debug('Reporting result to %r: %r', callback, result)
            callback(result)

        # Unblock any wait() calls
        self._result_arrived.set()

    @property
    def result(self):
        """Input from the user"""
        return self._result

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._parameters == other._parameters
        else:
            return NotImplemented

    def __repr__(self):
        arguments = []
        if self._parameters:
            arguments.append(', '.join((
                f'{k}={v!r}'
                for k, v in self._parameters.items()
            )))
        arguments.append(f'callbacks={self._callbacks}')
        arguments_string = ', '.join(arguments)
        return f'{type(self).__name__}({arguments_string})'


class RadioListPrompt(Prompt):
    """
    Pick one of two or more options

    :param options: Sequence of choices
    :param question: Question to show alongside the options or `None`
    :param focused: One of the `options` to focus initially or `None` to focus
        the first option
    """

    def __init__(self, *, callbacks=(), options, question=None, focused=None):
        super().__init__(
            callbacks=callbacks,
            options=options,
            question=question,
            focused=focused,
        )


class CheckListPrompt(Prompt):
    """
    Pick multiple of two or more options

    :param options: Sequence of choices
    :param question: Question to show alongside the options or `None`
    :param focused: One of the `options` to focus initially or `None` to focus
        the first option
    """

    def __init__(self, *, callbacks=(), options, question=None, focused=None):
        super().__init__(
            callbacks=callbacks,
            options=options,
            question=question,
            focused=focused,
        )


class TextPrompt(Prompt):
    """
    Input of arbitrary text

    :param question: Question to show alongside the options or `None`
    :param text: Prefilled text the user can delete or edit
    """

    def __init__(self, *, callbacks=(), question=None, text=''):
        super().__init__(callbacks=callbacks, question=question, text=text)



"""
Background workers to keep the UI responsive
"""

import asyncio
import enum
import functools
import multiprocessing
import pickle
import queue

from .. import errors

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MsgType(enum.Enum):
    """
    Enum that specifies the type of an IPC message (info, error, etc)
    """
    init = 'init'
    info = 'info'
    error = 'error'
    result = 'result'
    terminate = 'terminate'


class DaemonProcess:
    """
    :class:`multiprocessing.Process` abstraction with IPC

    Intended to offload heavy work (e.g. torrent creation) onto a different
    process. (Threads can still make the UI unresponsive because of the
    `GIL <https://wiki.python.org/moin/GlobalInterpreterLock>`_.)
    """

    def __init__(self, target, name=None, args=(), kwargs={},
                 init_callback=None, info_callback=None, error_callback=None,
                 result_callback=None, finished_callback=None):
        self._target = target
        self._name = name or target.__qualname__
        self._target_args = args
        self._target_kwargs = kwargs
        self._init_callback = init_callback
        self._info_callback = info_callback
        self._error_callback = error_callback
        self._result_callback = result_callback
        self._finished_callback = finished_callback

        self._ctx = multiprocessing.get_context('spawn')
        self._write_queue = self._ctx.Queue()
        self._read_queue = self._ctx.Queue()
        self._process = None
        self._process_finished = False
        self._read_queue_reader_task = None
        self._exception = None

    def start(self):
        """Start the process"""
        _log.debug('%s: Starting process: %r', self._name, self._target)
        target_wrapped = functools.partial(
            _target_process_wrapper,
            self._target,
            self._read_queue,
            self._write_queue,
        )
        self._process = self._ctx.Process(
            name=self._name,
            target=target_wrapped,
            args=self._make_args_picklable(self._target_args),
            kwargs=self._make_kwargs_picklable(self._target_kwargs),
        )
        self._process.start()
        self._read_queue_reader_task = asyncio.create_task(self._read_queue_reader())
        self._read_queue_reader_task.add_done_callback(self._handle_read_queue_reader_task_done)

    def _make_args_picklable(self, args):
        return tuple(self._get_picklable_value(arg) for arg in args)

    def _make_kwargs_picklable(self, kwargs):
        return {self._get_picklable_value(k): self._get_picklable_value(v)
                for k, v in kwargs.items()}

    def _get_picklable_value(self, value):
        # Try to pickle `value`. If it fails, try to instantiate `value` as one
        # of it's parent classes. Ignore first class (which is identical to
        # type(value)) and the last class, which should always be (`object`).
        types = type(value).mro()[1:-1]
        while True:
            try:
                pickle.dumps(value)
            # Python 3.9.1: This raises AttributeError, but the docs don't mention it.
            except Exception:
                if types:
                    value = types.pop(0)(value)
                else:
                    raise
            else:
                return value

    async def _read_queue_reader(self):
        while True:
            typ, msg = await asyncio.get_running_loop().run_in_executor(None, self._read_queue.get)

            if typ is MsgType.init:
                if self._init_callback:
                    self._init_callback(msg)

            elif typ is MsgType.info:
                if self._info_callback:
                    self._info_callback(msg)

            elif typ is MsgType.error:
                self._handle_error(msg)

            elif typ is MsgType.result:
                if self._result_callback:
                    self._result_callback(msg)
                break

            elif typ is MsgType.terminate:
                break

            else:
                raise RuntimeError(f'Unknown message type: {typ!r}')

    def _handle_read_queue_reader_task_done(self, task):
        # Catch exceptions from _read_queue_reader()
        try:
            task.result()

        except asyncio.CancelledError:
            _log.debug('%s: Cancelled _read_queue_reader()', self._name)

        except BaseException as e:
            _log.debug('%s: Caught _read_queue_reader() exception: %r', self._name, e)
            self._exception = e
            self._handle_error(e)

        finally:
            self._process_finished = True
            if self._finished_callback:
                try:
                    self._finished_callback()
                except BaseException as e:
                    self._exception = e
                    self._handle_error(e)

    def _handle_error(self, error):
        if isinstance(error, tuple):
            exception, traceback = error
            exception = errors.DaemonProcessError(exception, traceback)
        elif isinstance(error, BaseException):
            exception = error
        else:
            exception = errors.DaemonProcessError(errors.UpsiesError(error), '')

        if isinstance(exception, errors.DaemonProcessTerminated):
            return

        self._exception = exception
        if self._error_callback:
            try:
                self._error_callback(self._exception)
            except BaseException as e:
                self._exception = e
                raise
        else:
            raise self._exception

    def stop(self):
        """
        Stop the process

        This has no effect if the process is not alive.
        """
        if not self.is_finished:
            self.send(MsgType.terminate, None)

    def send(self, typ, value):
        """
        Send message to process if it is still alive

        If :attr:`is_alive` is `False`, do nothing.

        :param MsgType typ: General message category
        :param value: Message string, payload, sentinel, etc.
        """
        if self.is_finished:
            raise RuntimeError(f'Cannot send to finished process: {self._name}')
        else:
            self._write_queue.put((typ, value))

    @property
    def is_alive(self):
        """Whether :meth:`start` was called and the process has not finished yet"""
        if self._process:
            return self._process.is_alive()
        else:
            return False

    @property
    def is_finished(self):
        """Whether :meth:`start` was called and the process has finished"""
        return self._process_finished

    @property
    def exception(self):
        """Exception from `target` or callback, `None` if no exception was raised"""
        task = self._read_queue_reader_task
        if task and task.done() and task.exception():
            return task.exception()
        else:
            return self._exception

    async def join(self):
        """Block asynchronously until the process exits"""
        if self._process:
            _log.debug('%s: Joining process: %r', self._name, self._process)
            await asyncio.get_running_loop().run_in_executor(None, self._process.join)
            self._process = None

        if self._read_queue_reader_task:
            if not self._read_queue_reader_task.done():
                # Unblock self._read_queue()
                self._read_queue.put((MsgType.terminate, None))
                await self._read_queue_reader_task
            exc = self.exception
            if exc:
                _log.debug('Re-raising exception: %r', exc)
                raise exc


def _target_process_wrapper(target, write_queue, read_queue, *args, **kwargs):
    try:
        target(write_queue, read_queue, *args, **kwargs)
    except BaseException as e:
        # Because the traceback is not picklable, preserve it as a string before
        # sending it over the Queue
        import traceback
        traceback = ''.join(traceback.format_exception(type(e), e, e.__traceback__))
        write_queue.put((MsgType.error, (e, traceback)))
    finally:
        write_queue.put((MsgType.terminate, None))


def maybe_terminate(input_queue):
    """
    Read all queued items and raise :class:`~DaemonProcessTerminated` if there is a
    `MsgType.terminate`

    Put all other items back into the queue in the same order.
    """
    queued_items = read_input_queue_until_empty(input_queue)
    terminate = False
    for typ, msg in queued_items:
        if typ is MsgType.terminate:
            terminate = True
        else:
            input_queue.put((typ, msg))
    if terminate:
        raise errors.DaemonProcessTerminated('Terminated')


def read_input_queue(input_queue, default=(None, None)):
    """
    Return single `(<MsgType>, <payload data>)` tuple from `input_queue` or `default` if it is
    empty
    """
    try:
        typ, msg = input_queue.get_nowait()
    except queue.Empty:
        return default
    else:
        return typ, msg


NO_RETURN_VALUE = object()

def read_input_queue_key(input_queue, key):
    """
    Read :class:`dict` from `input_queue` and return its value for `key` if it exists

    The :class:`dict` must be queued as :class:`MsgType.info`.

    If `MsgType.terminate` is queued, raise :class:`~.DaemonProcessTerminated` instead of looking
    for `key`.

    Put all other items back into the queue in the same order.
    """
    return_value = NO_RETURN_VALUE
    terminate = False
    keep_looking = True
    items_to_requeue = []

    while keep_looking:
        queued_items = read_input_queue_until_empty(input_queue)

        # We must loop over all items here to check if there is a MsgType.terminate.
        for typ, msg in queued_items:
            if typ is MsgType.terminate:
                terminate = True
                keep_looking = False

            # It is important to only find the first matching key, otherwise we will lose all
            # leading matches if multiple matches are queued.
            elif (
                    return_value is NO_RETURN_VALUE
                    and typ == MsgType.info and isinstance(msg, dict)
                    and key in msg
            ):
                return_value = msg[key]
                keep_looking = False

            else:
                items_to_requeue.append((typ, msg))

    # Put all dequeued items back on the queue, except for the ones we processed.
    for item in items_to_requeue:
        input_queue.put(item)

    if terminate:
        raise errors.DaemonProcessTerminated('Terminated')
    else:
        return return_value


def read_input_queue_until_empty(input_queue):
    """
    Keep reading `input_queue` until it is empty

    Return the sequence of items that was read from the queue.

    Raise :class:`~DaemonProcessTerminated` if there is a `MsgType.terminate` queued.
    """
    items = []
    while True:
        try:
            typ, msg = input_queue.get(timeout=0.01)
        except queue.Empty:
            return tuple(items)
        else:
            if typ == MsgType.terminate:
                for item in items:
                    input_queue.put(item)
                raise errors.DaemonProcessTerminated('Terminated')
            else:
                items.append((typ, msg))



"""
Concrete :class:`~.TrackerJobsBase` subclass for BHD
"""

import functools
import io
import re

from ... import jobs, utils
from ..base import TrackerJobsBase
from . import metadata

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BhdTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.type_job,
            self.source_job,
            self.scene_check_job,
            self.tags_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
            self.rules_job,
            self.nfo_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "imdb_job") that were singled
        out by the user, e.g. with a CLI argument
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.release_name_job,
                # `release_name_job` doesn't depend on `imdb_job` (or any other
                # webdb), but we want the correct name, year, etc in the release
                # name.
                self.imdb_job,
            )
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def tmdb_job(self):
        tmdb_job = super().tmdb_job
        tmdb_job.no_id_ok = True
        return tmdb_job

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_category,
            autofinish=True,
            options=(
                ('Movie', '1'),
                ('TV', '2'),
            ),
            **self.common_job_args(),
        )

    _autodetect_category_map = {
        'Movie': lambda release_name: release_name.type is utils.release.ReleaseType.movie,
        'TV': lambda release_name: release_name.type in (utils.release.ReleaseType.season,
                                                         utils.release.ReleaseType.episode)
    }

    def autodetect_category(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting category: Approved release type: %r', approved_release_name.type)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(approved_release_name):
                return label

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_type,
            autofinish=True,
            options=(
                ('UHD 100', 'UHD 100'),
                ('UHD 66', 'UHD 66'),
                ('UHD 50', 'UHD 50'),
                ('UHD Remux', 'UHD Remux'),
                ('BD 50', 'BD 50'),
                ('BD 25', 'BD 25'),
                ('BD Remux', 'BD Remux'),
                ('2160p', '2160p'),
                ('1080p', '1080p'),
                ('1080i', '1080i'),
                ('720p', '720p'),
                ('576p', '576p'),
                ('540p', '540p'),
                ('DVD 9', 'DVD 9'),
                ('DVD 5', 'DVD 5'),
                ('DVD Remux', 'DVD Remux'),
                ('480p', '480p'),
                ('Other', 'Other'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    _autodetect_type_map = {
        # Directory trees / Images
        'UHD 100': lambda release_name: (
            release_name.source == 'BD100'
            and release_name.resolution == '2160p'
        ),
        'UHD 66': lambda release_name: (
            release_name.source == 'BD66'
            and release_name.resolution == '2160p'
        ),
        'UHD 50': lambda release_name: (
            release_name.source == 'BD50'
            and release_name.resolution == '2160p'
        ),
        'BD 50': lambda release_name: release_name.source == 'BD50',
        'BD 25': lambda release_name: release_name.source == 'BD25',
        'DVD 9': lambda release_name: release_name.source == 'DVD9',
        'DVD 5': lambda release_name: release_name.source == 'DVD5',

        # Remuxes
        'DVD Remux': lambda release_name: release_name.source.lower() == 'dvd remux',
        'UHD Remux': lambda release_name: (
            'remux' in release_name.source.lower()
            and release_name.resolution == '2160p'
        ),
        'BD Remux': lambda release_name: (
            'remux' in release_name.source.lower()
            and release_name.resolution == '1080p'
        ),

        # Encodes
        '2160p': lambda release_name: release_name.resolution == '2160p',
        '1080p': lambda release_name: release_name.resolution == '1080p',
        '1080i': lambda release_name: release_name.resolution == '1080i',
        '720p': lambda release_name: release_name.resolution == '720p',
        '576p': lambda release_name: release_name.resolution == '576p',
        '540p': lambda release_name: release_name.resolution == '540p',
        '480p': lambda release_name: release_name.resolution == '480p',
    }

    async def autodetect_type(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting type: Approved resolution: %r', approved_release_name.resolution)
        _log.debug('Autodetecting type: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(approved_release_name):
                return label

    @functools.cached_property
    def source_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_source,
            autofinish=True,
            options=tuple(metadata.sources),
            **self.common_job_args(),
        )

    async def autodetect_source(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting source: Approved source: %r', approved_release_name.source)
        for label, regex in metadata.sources.items():
            if regex.search(approved_release_name.source):
                return label

    release_name_english_title_before_original = True

    def _translate_video_format(video_format, release_name):
        # Do a (x|H(.))264 -> AVC/HEVC translation, but only for remux and BDMV.
        rn = release_name
        if re.search(r'(?:REMUX|BD25|BD50|BD66|BD100)', rn.source, flags=re.IGNORECASE):
            if re.search(r'^(?:x|H\.?)264$', video_format, flags=re.IGNORECASE):
                return 'AVC'
            elif re.search(r'^(?:x|H\.?)265$', video_format, flags=re.IGNORECASE):
                return 'HEVC'

    release_name_translation = {
        'source': {
            re.compile(r'(?i:remux)'): 'REMUX',
            re.compile(r'(?i:hybrid)'): 'HYBRID',
        },
        'video_format': _translate_video_format,
    }

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.playlists_job,
                self.mediainfo_job,
                self.bdinfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    def generate_description(self):
        screenshots = self._generate_description_screenshots()
        mediainfos = self._generate_description_mediainfos()

        description_parts = [
            '[center]\n'
            + screenshots
            + '\n[/center]'
        ]

        if mediainfos:
            # Include mediainfos if there are multiple (e.g. for IFO/VOB or if `document_all_videos`
            # is True)
            description_parts.append(
                '\n[center][h3]Mediainfo[/h3][/center]\n'
                + mediainfos
            )

        if self.promotion_bbcode:
            description_parts.append(f'\n\n\n{self.promotion_bbcode}')

        return ''.join(description_parts)

    def _generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished
        return self.make_screenshots_grid(
            screenshots=self.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        )

    def _generate_description_mediainfos(self):
        mediainfo_tags = []
        for video_filepath, mediainfo in self._description_mediainfos.items():
            filetitle = self.get_relative_file_path(video_filepath)
            mediainfo_tags.append(
                f'[hide={filetitle}][code]{mediainfo}[/code][/hide]'
            )
        return '\n'.join(mediainfo_tags)

    # Only generate mediainfo/bdinfo reports and screenshots for the first/main video file/playlist.
    document_all_videos = False

    @property
    def _description_mediainfos(self):
        # mediainfo_job should be disabled for BDMV releases. In that case, the BDInfo report is
        # passed normally in place of the mediainfo report. (See mediainfo_filehandle.)
        if self.mediainfo_job.is_enabled:
            assert self.mediainfo_job.is_finished

            # For VIDEO_TS releases, there should be one mediainfo report for a .IFO and another for
            # a .VOB file. The .IFO report should always be passed separately, but we include them
            # both in the description for ease of access.
            mediainfos_by_file = self.mediainfo_job.reports_by_file
            if len(mediainfos_by_file) >= 2:
                return mediainfos_by_file

        return {}

    @functools.cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            precondition=self.make_precondition('tags_job'),
            prejobs=(
                self.release_name_job,
                self.scene_check_job,
            ),
            text=self.generate_tags,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(),
        )

    async def generate_tags(self):
        assert self.release_name_job.is_finished
        assert self.scene_check_job.is_finished

        # Any additional tags separated by comma(s). (Commentary, 2in1, Hybrid,
        # OpenMatte, 2D3D, WEBRip, WEBDL, 3D, 4kRemaster, DualAudio, EnglishDub,
        # Personal, Scene, DigitalExtras, Extras)
        tags = []
        if 'WEBRip' in self.release_name.source:
            tags.append('WEBRip')
        elif 'WEB-DL' in self.release_name.source:
            tags.append('WEBDL')
        if 'Hybrid' in self.release_name.source:
            tags.append('Hybrid')
        if self.release_name.has_commentary:
            tags.append('Commentary')
        if self.release_name.has_dual_audio:
            tags.append('DualAudio')
        if 'Open Matte' in self.release_name.edition:
            tags.append('OpenMatte')
        if '2in1' in self.release_name.edition:
            tags.append('2in1')
        if '4k Remastered' in self.release_name.edition:
            tags.append('4kRemaster')
        if self.get_job_attribute(self.scene_check_job, 'is_scene_release'):
            tags.append('Scene')
        if self.options['personal_rip']:
            tags.append('Personal')

        # TODO: 2D3D
        # TODO: 3D
        # TODO: EnglishDub
        # TODO: DigitalExtras
        # TODO: Extras

        return '\n'.join(tags)

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'category_id': self.get_job_attribute(self.category_job, 'choice'),
            'type': self.get_job_attribute(self.type_job, 'choice'),
            'source': self.get_job_attribute(self.source_job, 'choice'),
            'imdb_id': self.get_job_output(self.imdb_job, slice=0),
            'tmdb_id': self.post_data_tmdb_id,
            'description': self.get_job_output(self.description_job, slice=0),
            'edition': self.post_data_edition,
            'custom_edition': self.options['custom_edition'],
            'tags': ','.join(self.get_job_output(self.tags_job, slice=0).split('\n')),
            'nfo': self.nfo_text,
            'pack': self.post_data_pack,
            'sd': self.post_data_sd,
            'special': self.post_data_special,
            'anon': '1' if self.options['anonymous'] else '0',
            'live': '0' if self.options['draft'] else '1',
        }

    @functools.cached_property
    def post_data_tmdb_id(self):
        # TMDb ID may be 0 for non-existing shows
        if self.tmdb_job.output:
            return self.get_job_output(self.tmdb_job, slice=0).split('/')[1]
        else:
            return 0

    @functools.cached_property
    def post_data_edition(self):
        # The edition of the uploaded release. (Collector, Director, Extended,
        # Limited, Special, Theatrical, Uncut or Unrated)
        edition = self.release_name.edition
        _log.debug('Approved edition: %r', edition)
        if "Collector's Edition" in edition:
            return 'Collector'
        elif "Director's Cut" in edition:
            return 'Director'
        elif 'Extended Cut' in edition:
            return 'Extended'
        elif 'Limited' in edition:
            return 'Limited'
        elif 'Special Edition' in edition:
            return 'Special'
        elif 'Theatrical Cut' in edition:
            return 'Theatrical'
        elif 'Uncut' in edition or 'Uncensored' in edition:
            return 'Uncut'
        elif 'Unrated' in edition:
            return 'Unrated'

    @property
    def post_data_pack(self):
        # The TV pack flag for when the torrent contains a complete season.
        # (0 = No TV pack or 1 = TV Pack). Default is 0
        if self.release_name.type is utils.release.ReleaseType.season:
            return '1'
        else:
            return '0'

    @property
    def post_data_sd(self):
        # The SD flag. (0 = Not Standard Definition, 1 = Standard Definition).
        # Default is 0
        try:
            height = int(self.release_name.resolution[:-1])
        except ValueError:
            return '0'
        else:
            return '1' if height < 720 else '0'

    @property
    def post_data_special(self):
        # The TV special flag for when the torrent contains a TV special. (0 =
        # Not a TV special, 1 = TV Special). Default is 0
        if (
                self.release_name.type is utils.release.ReleaseType.episode
                and self.options['special']
        ):
            return '1'
        else:
            return '0'

    # TODO
    # @property
    # def post_data_region(self):
    #     # The region in which the disc was released. Only for discs! (AUS,
    #     # CAN, CEE, CHN, ESP, EUR, FRA, GBR, GER, HKG, ITA, JPN, KOR, NOR,
    #     # NLD, RUS, TWN or USA)

    @property
    def mediainfo_filehandle(self):
        if self.is_bdmv_release:
            info = self.get_job_attribute(self.bdinfo_job, 'quick_summaries')[0]
        else:
            info = self.get_job_output(self.mediainfo_job, slice=0)
        return io.BytesIO(bytes(info, 'utf-8'))



import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit

from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ChoiceJobWidget(JobWidgetBase):

    is_interactive = True

    def setup(self):
        if self.job.multichoice:
            self._list_widget = widgets.CheckList(
                question=self.job.question,
                options=self.job.options,
                focused=self.job.focused,
                autodetected_indexes=self.job.autodetected_indexes,
                on_accepted=self._handle_accepted,
            )
        else:
            self._list_widget = widgets.RadioList(
                question=self.job.question,
                options=self.job.options,
                focused=self.job.focused,
                autodetected_index=(
                    self.job.autodetected_indexes[0]
                    if self.job.autodetected_indexes else
                    None
                ),
                on_accepted=self._handle_accepted,
            )
        self._activity_indicator = widgets.ActivityIndicator()
        self.job.signal.register('dialog_updated', self._handle_dialog_updated)

        self.job.signal.register('autodetecting', self._handle_autodetecting)
        self.job.signal.register('autodetected', self._handle_autodetected)

        # Immediately hide the radiolist. Otherwise it flashes into view when
        # the job is started only to become hidden when the autodetect function
        # is called a few milliseconds later.
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())

    def _handle_dialog_updated(self):
        self._list_widget.options = self.job.options
        self._list_widget.focused_index = self.job.focused_index
        if isinstance(self._list_widget, widgets.CheckList):
            self._list_widget.autodetected_indexes = self.job.autodetected_indexes
            self._list_widget.marked_indexes.update(self.job.autodetected_indexes)
        else:
            self._list_widget.autodetected_index = (
                self.job.autodetected_indexes[0]
                if self.job.autodetected_indexes else
                None
            )
        self.invalidate()

    def _handle_autodetecting(self):
        self._activity_indicator.enable()
        self.invalidate()

    def _handle_autodetected(self):
        self._activity_indicator.disable()
        self.invalidate()

    def _handle_accepted(self, choice):
        self.job.make_choice(choice)

    @functools.cached_property
    def runtime_widget(self):
        return HSplit(
            children=[
                ConditionalContainer(
                    content=self._activity_indicator,
                    filter=Condition(lambda: self._activity_indicator.active),
                ),
                ConditionalContainer(
                    content=self._list_widget,
                    filter=Condition(lambda: not self._activity_indicator.active),
                ),
            ],
        )


class TextFieldJobWidget(JobWidgetBase):

    is_interactive = True

    def setup(self):
        self._input_field = widgets.InputField(
            text=self.job.text,
            read_only=self.job.read_only,
            style='class:dialog.text',
            on_accepted=self._on_accepted,
        )
        self.job.signal.register('is_loading', self._handle_is_loading)
        self.job.signal.register('read_only', self._handle_read_only)
        self.job.signal.register('text', self._handle_text)

    def _handle_is_loading(self, is_loading):
        self._input_field.is_loading = is_loading
        self.invalidate()

    def _handle_read_only(self, read_only):
        self._input_field.read_only = read_only
        self.invalidate()

    def _handle_text(self, text):
        self._input_field.text = text
        self.invalidate()

    def _on_accepted(self, buffer):
        self.job.add_output(buffer.text)

    @functools.cached_property
    def runtime_widget(self):
        return self._input_field



import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit

from .. import widgets
from . import JobWidgetBase


class CustomJobWidget(JobWidgetBase):

    def setup(self):
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            extend_width=False,
        )
        # Deactivate activity indicator if we have output, which means the job is done.
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())
        self.job.signal.register('prompt', lambda _: self._activity_indicator.disable())
        self.job.signal.register('output', lambda _: self._activity_indicator.disable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())
        self.job.signal.register('indicate_activity', self._indicate_activity)

    def _indicate_activity(self, indicate_activity, text):
        if indicate_activity is not None:
            self._activity_indicator.active = indicate_activity
        if text is not None:
            self._activity_indicator.format = f'{{indicator}} {text}'
        self.invalidate()

    @functools.cached_property
    def runtime_widget(self):
        return HSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._activity_indicator.active),
                    content=self._activity_indicator,
                ),
            ],
        )
